
/*****************************************************************************
 *  The following codes compute the far fields based on the surface current  *
 *  obtained by emap5 			                                         *
 *									                       *
 *  Authors:  Dr. Mohammad W. Ali,   Yun  Ji,  Dr. Todd. H. Hubing           *
 *  Version:  1.1							                       *
 *  Last updated:  Dec 16, 1998   				                       *
 *****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "complex.c"
#include "alloc.c"
#include "util.c"


#ifndef M_PI  
#define M_PI 3.1415926
#endif

#define TotQuadPoint 7


/***********************  Function Prototype  *************************/
void   GetGaussQuadData();
void   ComputeGaussQuadPoint(int, int *, double *);
void   ReadInputFile();
void   ComputeNonSingul(int *, double, double, double, double, double);
void   ComputeFarEField(double *, double, double, double, double);
double Sign(int);


/***********************    GLOBAL VARIABLES   ************************/

int
   *InnerEdgeStat,
   *BoundEdgeStat,
   **PlusTrngleDetect,  *PlusTrngleIndex,
   **MinusTrngleDetect, *MinusTrngleIndex,
   **TrngleNode, **TrngleEdge, *BoundEdgeStat,
   **GlobalEdgeEnds,
   **TetEdge, **TetNode,
   TotEdgeNum,
   TotTetElement,
   TotNodeNum,
   TotTrngleNum,
   TotBoundEdgeNum,
   DielBoundEdgeNum,
   TotInnerEdgeNum,
   TotISourceEdgeNum,
   HybrdBoundEdgeNum;


double 
    WaveNumber, WaveLength, ImpeDance, FreeSpaceVel, AbsPermeable, AbsPermitt,
    OperateFreq,
    Wght[7], Qpnt[7][3], 
    RadDisObserv,
    **NodeCord;

complex
    InoPQ, IxiPQ, IetaPQ, IzetaPQ,  
    CFactor,
    CFactor1,  
    *Epsilon,
    *MVector,
    *JVector;

FILE *InF1,*InF2,*OutF;



main(int argc, char **argv)
{
    int IEdge, JEdge;
    double ThetaObserv1, ThetaObserv2, ThetaStep, 
           PhiObserv1, PhiObserv2, PhiStep, 
           ThetaObserv, PhiObserv,
           sth,sph,cth,cph, ObservPoint[3];  

    if (argc!=4)
    {
       fprintf(stderr,"Usage:far mesh SurfaceFieldFile OutputFile\n");
       exit(1);
    }

    if ((InF1  = fopen(argv[1],"r"))==NULL)
    {
       fprintf(stderr,"Can't open MeshFile:%s\n",argv[1]);
       exit(1);
    }

    if ((InF2 = fopen(argv[2],"r"))==NULL) 
    {
       fprintf(stderr,"Can't open SurfaceFieldFile:%s\n",argv[2]);
       exit(1) ;
    }

    if ((OutF = fopen(argv[3],"w"))==NULL)
    {
       fprintf(stderr,"Can't open OutputFile:%s\n",argv[3]);
       exit(1) ;
    }   

 
    GetGaussQuadData();
    ReadInputFile();
  
    FreeSpaceVel  = 3.0E+08;
    AbsPermeable  = 1.25663706144E-06;
    AbsPermitt    = 8.8542E-12;
    WaveLength = (FreeSpaceVel/(OperateFreq*1.0E+06));
    WaveNumber = 2.0*M_PI /WaveLength;
    ImpeDance  = WaveNumber/(2.0*M_PI*OperateFreq*1.0E+06*AbsPermitt);
    CFactor = Real_Mul(-1.0, COMplex_Cmplx(0.0,(WaveNumber*ImpeDance)));
           

    /* calculate the far field observed at (RadDisObserv,Theta,Phi)  */

    printf("Enter the observing distance(in wavelength, >=20): \t");
    scanf("%lf",&RadDisObserv);
    RadDisObserv *= WaveLength;

    printf("\nEnter Theta interval:\n\tStart Angle(in degree,>=0): \t");
    scanf("%lf", &ThetaObserv1);
    printf("\tEnd Angle(in degree, <=180): \t");
    scanf("%lf", &ThetaObserv2);
    printf("\tTheta step (!=0):\t");
    scanf("%lf", &ThetaStep);
 
    if (ThetaStep==0.0) 
    {
        fprintf(stderr, "Step can't be zero\n");
        exit(1);
    } 

    printf("\nEnter Phi interval:\n\tStart Angle(in degree, >=0): \t");
    scanf("%lf", &PhiObserv1);
    printf("\tEnd Angle(in degree,<=180): \t");
    scanf("%lf", &PhiObserv2);
    printf("\tPhi step(!=0):\t");
    scanf("%lf", &PhiStep);
 
    if (PhiStep==0.0) 
    {
        fprintf(stderr, "Step can't be zero\n");
        exit(1);
    }
     
    printf("\nPlease wait ......\n");
    fprintf(OutF, "Phi\t\tTheta\t\tE_Phi\t\tE_Theta\n");

    for(IEdge=0; IEdge<=(PhiObserv2-PhiObserv1)/PhiStep; IEdge++)
    {
        PhiObserv = PhiObserv1+ IEdge*PhiStep ;
        for(JEdge=0;JEdge<=(ThetaObserv2-ThetaObserv1)/ThetaStep;JEdge++)
        {
            ThetaObserv = ThetaObserv1+ JEdge*ThetaStep; 
            fprintf(OutF, "%lf\t%lf\t",  PhiObserv, ThetaObserv );  
            sth = sin(ThetaObserv * (M_PI/180.0));
            cth = cos(ThetaObserv * (M_PI/180.0));
            sph = sin(PhiObserv   * (M_PI/180.0));
            cph = cos(PhiObserv   * (M_PI/180.0));
            ObservPoint[0] = RadDisObserv * (sth * cph); 
            ObservPoint[1] = RadDisObserv * (sth * sph); 
            ObservPoint[2] = RadDisObserv * ( cth);
            ComputeFarEField(ObservPoint,cth,cph,sth,sph);
        }   /* for(JEdge */
    }  /* for(IEdge */

}  /* end of main()  */



/******************* ReadInputFile() *******************/
void ReadInputFile()
{
    int i,j ; 
    char buff[100], c;
 
    /* read the global information table */
    fscanf(InF1, " %c", &c);
    if (c!='#') 
    {
       fprintf(stderr, "Check your global information table\n");
       exit(1);
    }
    fgets(buff, sizeof(buff)-1, InF1);

    fscanf(InF1,"%d",&TotEdgeNum);
    fscanf(InF1,"%d",&TotTetElement);
    fscanf(InF1,"%d",&TotNodeNum);
    fscanf(InF1,"%d",&TotTrngleNum);
    fscanf(InF1,"%d",&TotBoundEdgeNum);
    fscanf(InF1,"%d",&TotInnerEdgeNum);
    fscanf(InF1,"%d",&DielBoundEdgeNum);
    fscanf(InF1,"%d",&HybrdBoundEdgeNum);

    /* read the global node table */
    fscanf(InF1, " %c", &c);
    if (c!='#') 
    {
       fprintf(stderr, "Check your global node table\n");
       exit(1);
    }
    fgets(buff, sizeof(buff)-1, InF1);
    NodeCord = double_Matrix(TotNodeNum,3);
    for(i=0;i<TotNodeNum;i++) 
        for(j=0;j<=2;j++)
            fscanf(InF1,"%lf",&NodeCord[i][j]);

    /* read the global edge table */
    fscanf(InF1, " %c", &c);
    if (c!='#') 
    {
       fprintf(stderr, "Check your global edge table\n");
       exit(1);
    }
    fgets(buff, sizeof(buff)-1, InF1);
    GlobalEdgeEnds= INT_Matrix(TotEdgeNum,6);
    for(i=0;i<TotEdgeNum;i++)
        for(j=0;j<6;j++)
            fscanf(InF1,"%d",&GlobalEdgeEnds[i][j]);


    /* read the triangle table */
    fscanf(InF1, " %c", &c);
    if ( c!='#' ) 
    {
       fprintf(stderr, "Check your triangle table\n");
       exit(1);
    }
    fgets(buff, sizeof(buff)-1, InF1);
    TrngleNode = INT_Matrix(TotTrngleNum,3);
    TrngleEdge = INT_Matrix(TotTrngleNum,3);
    for(i=0;i<TotTrngleNum;i++)
    {
        for(j=0;j<=2;j++) fscanf(InF1,"%d",&TrngleNode[i][j]);
        for(j=0;j<=2;j++) fscanf(InF1,"%d",&TrngleEdge[i][j]);
    }
 
  
    /* read tetrahedron table */
    fscanf(InF1, " %c", &c);
    if (c!='#') 
    {
       fprintf(stderr, "Check your tetrahedron table\n");
       exit(1);
    }
    fgets(buff, sizeof(buff)-1, InF1);
    TetNode = INT_Matrix(TotTetElement,4);
    Epsilon =CMPLX_Vector(TotTetElement);
    for(i=0;i<TotTetElement;i++)
    {
        for(j=0;j<=3;j++)
            fscanf(InF1,"%d",&TetNode[i][j]);
        fscanf(InF1,"%lf",&Epsilon[i].x);
        fscanf(InF1,"%lf",&Epsilon[i].y);
    }

 
    TetEdge = INT_Matrix(TotTetElement,6);
    for(i=0;i<TotTetElement;i++) for(j=0;j<=5;j++)
        fscanf(InF1,"%d",&TetEdge[i][j]);


    /* read plus/minus triangle table */    
    fscanf(InF1, " %c", &c);
    if (c!='#') 
    {
       fprintf(stderr, "Check your plus/minus triangle table\n");
       exit(1);
    }
    fgets(buff, sizeof(buff)-1, InF1);
    PlusTrngleDetect = INT_Matrix(TotEdgeNum,3);
    PlusTrngleIndex = INT_Vector(TotEdgeNum);
    MinusTrngleDetect = INT_Matrix(TotEdgeNum,3);
    MinusTrngleIndex = INT_Vector(TotEdgeNum);
    for(i=0;i<TotEdgeNum;i++)
    {
        fscanf(InF1,"%d",&PlusTrngleIndex[i]);
        for(j=0;j<PlusTrngleIndex[i];j++)
            fscanf(InF1,"%d",&PlusTrngleDetect[i][j]);
        fscanf(InF1,"%d",&MinusTrngleIndex[i]);
        for(j=0;j<MinusTrngleIndex[i];j++)
            fscanf(InF1,"%d",&MinusTrngleDetect[i][j]);
    }
 
    fscanf(InF1, " %c", &c);
    if (c!='#') 
    {
       fprintf(stderr, "Check your inner edge table\n");
       exit(1);
    }
    fgets(buff, sizeof(buff)-1, InF1);
    InnerEdgeStat = INT_Vector(TotInnerEdgeNum);
    for(i=0;i<TotInnerEdgeNum;i++)
        fscanf(InF1,"%d",&InnerEdgeStat[i]);


    /* read boundary edge table */
    fscanf(InF1, " %c", &c);
    if (c!='#') 
    {
       fprintf(stderr, "Check your boundary edge table\n");
       exit(1);
    }
    fgets(buff, sizeof(buff)-1, InF1);
    BoundEdgeStat = INT_Vector(TotBoundEdgeNum);
    for(i=0;i<TotBoundEdgeNum;i++)
        fscanf(InF1,"%d",&BoundEdgeStat[i]);

    fscanf(InF1, " %c", &c);
    if (c!='#') 
    {
       fprintf(stderr, "Check your source information table\n");
       exit(1);
    }
    fgets(buff, sizeof(buff)-1, InF1);
    fscanf(InF1," %c", &c);  /* skip the type */
    fscanf(InF1,"%lf",&OperateFreq);  /* obtain the working frequency */

  
    MVector = CMPLX_Vector(TotBoundEdgeNum);
    JVector = CMPLX_Vector(TotBoundEdgeNum);


    for(j=0;j<HybrdBoundEdgeNum; j++)
        if (fscanf(InF2,"%lf %lf",&MVector[j].x,&MVector[j].y)!=2)
        {
           fprintf(stderr, "Surface field file damaged\n");
           exit(1);
        }
                                
    for(j=0; j<TotBoundEdgeNum-HybrdBoundEdgeNum; j++)
        MVector[j+HybrdBoundEdgeNum]=COMplex_Cmplx(0.0, 0.0);

    for(j=0;j<TotBoundEdgeNum;j++) 
        if (fscanf(InF2,"%lf %lf",&JVector[j].x,&JVector[j].y)!=2)
        {
           fprintf(stderr, "Surface field file damaged\n");
           exit(1);
        }

}



/********************** ComputeFarEField() *******************************/
void ComputeFarEField(double *ObservPoint, double cth, double cph, 
                        double sth, double sph)
{

    int IEdge, SrcEdgeNum, SrcNodeNum, EdgeEnd1, EdgeEnd2, RowNum,
        SourceEdge, SourceTrngle, TrngleNodeArray[3] ;
    double tmp, Sgn;
    complex ETheta, EPhi, ATheta, APhi, FTheta, FPhi,
            AVectorX, AVectorY, AVectorZ, AVectorPQ[3],
            FVectorX, FVectorY, FVectorZ,
            V_x,  V_y, V_z;

    ETheta = COMplex_Null();   EPhi =  COMplex_Null();
    AVectorX = COMplex_Null(); FVectorX = COMplex_Null();
    AVectorY = COMplex_Null(); FVectorY = COMplex_Null();
    AVectorZ = COMplex_Null(); FVectorZ = COMplex_Null();

    for(SourceTrngle=0;SourceTrngle<TotTrngleNum;SourceTrngle++)
    {

        InoPQ = COMplex_Null() ; 
        IetaPQ  = COMplex_Null() ;
        IxiPQ = COMplex_Null() ; IzetaPQ = COMplex_Null() ;

        for(IEdge=0;IEdge<=2;IEdge++) 
            TrngleNodeArray[IEdge] = TrngleNode[SourceTrngle][IEdge]-1;

        ComputeNonSingul(TrngleNodeArray,cth,cph,sth,sph,RadDisObserv);

        V_x = COMplex_Add2(Real_Mul(NodeCord[TrngleNodeArray[0]][0],  IxiPQ),
                   Real_Mul(NodeCord[TrngleNodeArray[1]][0], IetaPQ),
                   Real_Mul(NodeCord[TrngleNodeArray[2]][0],IzetaPQ));
        V_y = COMplex_Add2(Real_Mul(NodeCord[TrngleNodeArray[0]][1],  IxiPQ),
                   Real_Mul(NodeCord[TrngleNodeArray[1]][1], IetaPQ),
                   Real_Mul(NodeCord[TrngleNodeArray[2]][1],IzetaPQ));
        V_z = COMplex_Add2(Real_Mul(NodeCord[TrngleNodeArray[0]][2],  IxiPQ),
                   Real_Mul(NodeCord[TrngleNodeArray[1]][2], IetaPQ),
                   Real_Mul(NodeCord[TrngleNodeArray[2]][2],IzetaPQ));

        for(SourceEdge=0;SourceEdge<=2;SourceEdge++)
        {
            double  SrcTrngleEdgeLen;

            SrcEdgeNum = abs(TrngleEdge[SourceTrngle][SourceEdge]);
            SrcNodeNum = TrngleNode[SourceTrngle][SourceEdge] - 1;
            Sgn        = Sign(TrngleEdge[SourceTrngle][SourceEdge]);
            EdgeEnd1   = GlobalEdgeEnds[SrcEdgeNum-1][0]-1;
            EdgeEnd2   = GlobalEdgeEnds[SrcEdgeNum-1][1]-1;
            SrcTrngleEdgeLen = VTXmag(NodeCord[EdgeEnd1],NodeCord[EdgeEnd2]);

            for(IEdge=0;IEdge<TotBoundEdgeNum;IEdge++)
            {
                if (BoundEdgeStat[IEdge]==SrcEdgeNum) 
                {
                    RowNum=IEdge;
                    break;
                } 

                else  RowNum=-1;
            }

            if (RowNum!=-1) 
            {
  
                AVectorPQ[0] = COMplex_Sub(V_x,  
                          Real_Mul(NodeCord[SrcNodeNum][0],InoPQ));
                AVectorPQ[1] = COMplex_Sub(
                          V_y,Real_Mul(NodeCord[SrcNodeNum][1],InoPQ));
                AVectorPQ[2] = COMplex_Sub(V_z, 
                          Real_Mul(NodeCord[SrcNodeNum][2],InoPQ));

                CFactor1= Real_Mul((Sgn * SrcTrngleEdgeLen),JVector[RowNum]);

                AVectorX = COMplex_Add(AVectorX,
                             COMplex_Mul( CFactor1,AVectorPQ[0]));
                AVectorY = COMplex_Add(AVectorY,
                             COMplex_Mul( CFactor1,AVectorPQ[1]));
                AVectorZ = COMplex_Add(AVectorZ,
                             COMplex_Mul( CFactor1,AVectorPQ[2]));

                CFactor1 = Real_Mul((Sgn * SrcTrngleEdgeLen),MVector[RowNum]);

                FVectorX = COMplex_Add(FVectorX,
                             COMplex_Mul( CFactor1,AVectorPQ[0]));
                FVectorY = COMplex_Add(FVectorY,
                              COMplex_Mul( CFactor1,AVectorPQ[1]));
                FVectorZ = COMplex_Add(FVectorZ,
                             COMplex_Mul( CFactor1,AVectorPQ[2]));

            }  /* if(RowNum!=1) */
        }
    }

    ATheta = COMplex_Sub( COMplex_Add(Real_Mul((cth*cph),AVectorX),
                                       Real_Mul((cth*sph),AVectorY)),
                          Real_Mul(sth,AVectorZ));
                     
    APhi =   COMplex_Sub( Real_Mul( cph , AVectorY),
                     Real_Mul( sph , AVectorX) ) ;
    FTheta = COMplex_Sub( COMplex_Add(Real_Mul((cth*cph),FVectorX),
                                      Real_Mul((cth*sph),FVectorY)),
                          Real_Mul( sth,FVectorZ) );
                     
                     
    FPhi = COMplex_Sub(Real_Mul(cph,FVectorY), Real_Mul(sph,FVectorX)) ;
 

    FTheta = Real_Mul(1.0/ImpeDance, FTheta);
    FPhi   = Real_Mul(1.0/ImpeDance, FPhi);

    ETheta = COMplex_Mul(CFactor,COMplex_Add(ATheta,FPhi));
    EPhi   = COMplex_Mul(CFactor,COMplex_Sub(APhi,FTheta));


    tmp=1.0/(4.0*M_PI*RadDisObserv);


    ETheta = Real_Mul(tmp,ETheta ); 
    EPhi   = Real_Mul(tmp, EPhi );

    /********E filed computation ***********/
    fprintf(OutF, "%e\t%e\n", COMplex_Abs(EPhi),COMplex_Abs(ETheta)); 

}   /*  end of ComputeFarEField( )  */
 
 

/********************* ComputeNonSingul() **********************/
void ComputeNonSingul(int *TrngleNode, double cth, double cph, double sth,  
                      double sph, double ObsDist)
{
    int QuadPoint;
    double Rcp,SrcPoint[3]; 
    complex V1, V2, Eprj;

   
    for(QuadPoint=0;QuadPoint<=TotQuadPoint-1;QuadPoint++)
    {
        ComputeGaussQuadPoint(QuadPoint,TrngleNode,SrcPoint);
        Rcp = (SrcPoint[0]*cph + SrcPoint[1]*sph)*sth + SrcPoint[2]*cth;
   
        V1 = COMplex_Cmplx(0.0,(Rcp*WaveNumber));
        V1 = COMplex_Expon(1.0,V1);
   
        /* Following codes are quoted by Ali*/
        V2 = COMplex_Expon(-1.0,COMplex_Cmplx(0.0,(ObsDist*WaveNumber)));
        V1 = COMplex_Mul(V1,V2);
 
        Eprj = Real_Mul(Wght[QuadPoint],V1);
        InoPQ = COMplex_Add(InoPQ, Eprj);
        IxiPQ = COMplex_Add(IxiPQ,  Real_Mul(Qpnt[QuadPoint][0],Eprj) );
        IetaPQ = COMplex_Add(IetaPQ, Real_Mul(Qpnt[QuadPoint][1],Eprj) );
    }

    IzetaPQ    = COMplex_Sub(InoPQ,COMplex_Add(IxiPQ ,IetaPQ) );

}
 
 

/************** ComputeGaussQuadPoint() ********************************/
void ComputeGaussQuadPoint(int QuadPoint, int *TrngleNode, double *SrcPointCol)
{
     SrcPointCol[0] =   Qpnt[QuadPoint][0]*NodeCord[TrngleNode[0]][0]
                      + Qpnt[QuadPoint][1]*NodeCord[TrngleNode[1]][0]
                      + Qpnt[QuadPoint][2]*NodeCord[TrngleNode[2]][0];

     SrcPointCol[1] =   Qpnt[QuadPoint][0]*NodeCord[TrngleNode[0]][1]
                      + Qpnt[QuadPoint][1]*NodeCord[TrngleNode[1]][1]
                      + Qpnt[QuadPoint][2]*NodeCord[TrngleNode[2]][1];

     SrcPointCol[2] =   Qpnt[QuadPoint][0]*NodeCord[TrngleNode[0]][2]
                      + Qpnt[QuadPoint][1]*NodeCord[TrngleNode[1]][2]
                      + Qpnt[QuadPoint][2]*NodeCord[TrngleNode[2]][2];

}   /* ComputeGaussQuadPoint( ) */




/***************** GetGaussQuadData() ***************************************/
void GetGaussQuadData()
{
    int QuadPoint;

    Wght[0]=0.112500000000000; 
    Qpnt[0][0]= Qpnt[0][1]=0.333333333333333;    

    Wght[1]=0.062969590272413; 
    Qpnt[1][0]=0.797426985353087; 
    Qpnt[1][1]=0.101286507323456;

    Wght[2]=0.062969590272413; 
    Qpnt[2][0]=0.101286507323456; 
    Qpnt[2][1]=0.797426985353087;

    Wght[3]=0.062969590272413;
    Qpnt[3][0]= Qpnt[3][1]=0.101286507323456; 

    Wght[4]=0.066197076394253;
    Qpnt[4][0]= Qpnt[4][1]=0.470142064105115; 
 
    Wght[5]=0.066197076394253;
    Qpnt[5][0]=0.470142064105115; 
    Qpnt[5][1]=0.059715871789770;

    Wght[6]=0.066197076394253; 
    Qpnt[6][0]=0.059715871789770; 
    Qpnt[6][1]=0.470142064105115;

    for(QuadPoint=0;QuadPoint<=TotQuadPoint-1;QuadPoint++)
        Qpnt[QuadPoint][2] = 1.0 - Qpnt[QuadPoint][0] - Qpnt[QuadPoint][1];

}  /* end of GetGaussQuadData() */



double Sign(int Value)
{
    return (Value>0 ? 1.0:-1.0);
}

