/*****************************************************************************
 *  The following codes read a sif file as the input file, then generate     *
 *  the mesh data, source information and output requirements for emap5.     *
 *                        						                 *
 *  Authors:  Dr. Mohammad W. Ali,   Yun  Ji,  Dr. Todd. H. Hubing           *
 *  Version:  1.1							                       *
 *  Last updated:  Dec 16, 1998   				                       *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "complex.c"
#include "alloc.c"
#include "util.c"

typedef struct {
    int x1,y1,z1,x2,y2,z2;   /* two corner points of the conductor */
    double   DelX, DelY, DelZ;  /* step size */
} Cond;


typedef struct{
    double  start;  /* start point */
    double  end;    /* end point */
    double  step;   /* step */
} CellUnit;


typedef struct{
    int x;        
    int y;
} XPoint;



typedef struct{
   XPoint start;
   XPoint end;
}  RectAngle;



/******************   FUNCTION PROTOTYPE  ************************/
void AllocateMemory();
void AssignBoxCoordinates();
void AssignHexHedronEdgeNum();
void AssignHexHedronNodeNum();
void ComputeEdgeMultiplicity();
void ComputeParameters();
void ComputeTrngleCentroid(int, double **, int **, double **);
void DetectBoundEdgeStatus();
void DetectInnerEdgeStatus();
void DetectISourceEdges();
void DetectVSourceEdges();
void DiscretizeFEMVolume();
void DiscretizeExterPECPatch();
void FindInterMetalEdges();
void FindTetraHedronFaceNodes();
void GlobalEdgeEndsDiv1();
void GlobalEdgeEndsDiv2();
int  Intersection(RectAngle, RectAngle, RectAngle *, int *, RectAngle *);
int  Max(int, int);
int  Min(int, int);
void ModifyBoundFaces(int );
void PrintOutput();
void ReadSIFFile();
void TetraEdgeDiv1();


/******************** GLOBAL VARIABLES **********************/

int  DetectCondBound(Cond);
void DetectInnerCond(Cond);


CellUnit CellUnitX[10], CellUnitY[10], CellUnitZ[10];
int   CellUnitXNum=0;
int   CellUnitYNum=0;
int   CellUnitZNum=0;

Cond conductor[20];
int CondNum;


int flag_default_out=0; /* flag used to determine which kind of
                                           output users want */

int output_num;  /* how many times "output" are used, maxium number is 10 */ 
int output_x[2][10], 
    output_y[2][10], 
    output_z[2][10]; /* store information for  each "output" */

char output_file[10][20];  /* file to store the output data */
char output_axis[10];  /* only edges parallel to the designated axis 
                              wil be printed out */

char default_out_file[20]; /* file to store default output data */

double E_Theta, E_Phi, K_Theta, K_Phi, PlaneWaveE_Mag;


int JunctionNum=0;
/* maxium junction edge allowed is 100. 
   Junction[][0] ---- the hybrid edge number 
   Junction[][1] ---- the other external metal edge with same coordinates 
                      with Junction[][0]. these two form a juction      
   Junction[][2]---- the T++ or T-- triangle associate with these Junction */

int Junction[100][3];



/*  store the resistor edges */
int ResisterLocX[20][2];
int ResisterLocY[20][2];
int ResisterLocZ[20][2];
int ResisterNum;
double ResisterValue[20];


char  In_FileName[50];

int
    **HexNode,
    **HexEdgeNum,
    **TetTrngleNode,
    *TetBoundFaceTetHedronNum,
    *TetBoundFaceFaceNum,
    **TrngleEdge,
    **TrngleNode,
    **TetNode,
    **FaceNode,
    **FaceEdge,
    *EdgeFound,
    *MultFound,
    **PlusTrngleDetect,
    *PlusTrngleIndex,
    **MinusTrngleDetect,
    LastMultEdgeFoundNum,
    LastEdgeFoundNum,
    MetalInEdgeNum,
    MetalBoundEdgeNum,
    DipolWid[10],DipolLen[10],
    DielBoundEdgeNum,
    ExtEdge1,ExtEdge2,
    ExtEdge3,ExtEdge4,
    ExtEdge5,ExtEdge6,
    ImpEdge1,ImpEdge2,
    InnEdge1,InnEdge2,
    *MinusTrngleIndex,
    **TetEdge,
    *EdgeStat,
    *ISourceEdgeStat,
    *ForcdStat,
    TotISourceEdgeNum,
    *MetalEdgeStat,
    *InnerEdgeStat,
    *BoundEdgeStat,
    **GlobalEdgeEnds
;


int PatchNum,FrontWall,BackWall,BottomWall;
int TopWall,RightWall,LeftWall;

 
double
    VsourceMag,
    OperateFreq,
    CellDimension,
    **MediaPerm,
    **TrngleCenTroid,
    *TrngleArea,
    **TrngleNormal,
    ** NodeCord,
    *ISourceEdgeMag;



int CubeTrngleNum,CubeNumNode,CubeNumEdge,
    FrontEdge,FrontNode,BackEdge,BackNode;

int 
    IEdge,
    JEdge,
    VVal,II,KK,
    TotBoundEdgeNum,
    FExtDipolLen,
    BExtDipolLen,
    TotTrngleNum,
    TotDipolTrngleNum[10],
    TotCubeTrngleNum,
    HybrdBoundEdgeNum,
    TotMetalTrngleNum,
    ExtDipolNodeStart,
    ExtDipolEdgeStart,
    ExtDipolTrngleStart,
    TotInnerEdgeNum,
    Xcount,Ycount,Zcount,
    PLen,
    start, TotNumHexHedron,
    GlobalTetFaceNum,
    LocalTetFaceNum,
    i,j,k,ii,jj,kk,Count,
    TotTetFaceNum=0,
    XdiM,YdiM,ZdiM,
    TotEdgeNum,
    TotDipolNumEdge[10],
    TotNodeNum,
    TotDipolNumNode[10],
    TotCubeNumEdge,
    TotCubeNumNode,
    HexNum,con,
    PermConst,
    nnode,
    TotTetElement;

int 
    XBoundMinDim,YBoundMinDim,ZBoundMinDim,
    XBoundMaxDim,YBoundMaxDim,ZBoundMaxDim,
    XDielmat[2][10],YDielmat[2][10],ZDielmat[2][10],
    XintEsrc[2][10],YintEsrc[2][10],ZintEsrc[2][10],
    XextVsrc[2][10],YextVsrc[2][10],ZextVsrc[2][10] ,
    Xintcond[2][10],Yintcond[2][10],Zintcond[2][10],
    Xbndcond[2][10],Ybndcond[2][10],Zbndcond[2][10],
    Xextcond[2][10],Yextcond[2][10],Zextcond[2][10],
    XmaxWall,YmaxWall,ZmaxWall,
    XminWall,YminWall,ZminWall,
    DelXintcond[10],DelYintcond[10],DelZintcond[10],
    DelXbndcond[10],DelYbndcond[10],DelZbndcond[10],
    DelXextcond[10],DelYextcond[10],DelZextcond[10],
    XBoundDim[2],YBoundDim[2],ZBoundDim[2],
    VSourceEdge[20],
    VSourceNum=0,
    DielmatNum=0,
    IntEsrcNum=0,
    ExtVsrcNum=0,
    IntcondNum=0,
    BndcondNum=0,
    ExtcondNum=0,
    Max_X=0,
    Max_Y=0,
    Max_Z=0,
    Min_X=0,
    Min_Y=0,
    Min_Z=0;


double  VSourceMag[20],ExtVsrcMag[10] ;


char SourceType;
char   IntEsrcAxis[10];   /* store E source direction */
double IntEsrceMag[10];   /* store Esource magnetude*/


char PolInc[1];
double ThetaInc,PhiInc;
double Epsilon[10],Sigma[10];


int Dipolplane_XY[10],Dipolplane_YZ[10],Dipolplane_ZX[10];

char Out_FileName0[20],
     Out_FileName1[20],
     Out_FileName2[20],
     Out_FileName3[20];

FILE *InF, *OutF_0, *OutF_1, *OutF_2, *OutF_3, *OutFilePntr;



int main(int argc, char **argv)
{
    char buff[10];

    if( argc != 3 ) 
    {
        fprintf(stderr,"Usage: sift5 inputfile outputfile\n");
        exit(1);
    }

    if (strlen(argv[1])<=4) 
    {
       fprintf(stderr, "Error: Use *.sif as the input file\n");
       exit(1);
    }
       
    /* find out the surfix of the input file */
    strcpy(buff, argv[1]+strlen(argv[1])-4 );
    if (strcmp(buff, ".sif")!=0) 
    {
       fprintf(stderr, "Error: Use *.sif as the input file\n");
       exit(1);
    }

    if (strlen(argv[2])<=3) 
    {
       fprintf(stderr, "Error: Use *.in as the output file\n");
       exit(1);
    }
       
    /* find out the surfix of the output file */
    strcpy(buff, argv[2]+strlen(argv[2])-3 );
    if (strcmp(buff, ".in")!=0) 
    {
       fprintf(stderr, "Error: Use *.in as the input file\n");
       exit(1);
    }
      
    if( (InF = fopen(argv[1], "r"))== NULL ) 
    {
        fprintf(stderr, " Error: Can't open the sif file.\n");
        exit(1);
    }
   

    if ( (OutFilePntr = fopen(argv[2],"w"))==NULL) 
    {
         fprintf(stderr, " Error: Can't open output file.\n");
	 exit(1);
    }

 
    ReadSIFFile();
    ComputeParameters();
    AllocateMemory();
    AssignBoxCoordinates();
    AssignHexHedronNodeNum();
    AssignHexHedronEdgeNum();

    DetectInnerEdgeStatus();
    DetectBoundEdgeStatus();

    DiscretizeFEMVolume();

    DiscretizeExterPECPatch();
 
    FindInterMetalEdges();  
    
    ComputeEdgeMultiplicity();


    /* initialize ForcdStat[] */
    ForcdStat = INT_Vector(TotEdgeNum);

    for(ii=0; ii<TotEdgeNum; ii++)
        ForcdStat[ii]=0;   /* all are set as "not forcd" */

    ISourceEdgeStat=INT_Vector(TotEdgeNum);

    /* initilize ISourceEdgeMag */
    ISourceEdgeMag=double_Vector(TotEdgeNum);

    for(ii=0;ii<TotEdgeNum; ii++)
        ISourceEdgeMag[ii]=0.0;

  
    if( SourceType=='I' ) DetectISourceEdges();
   
    /* to build the inner-edge-table */
    kk=0;
    for(ii=0;ii<TotEdgeNum;ii++)
        if(EdgeStat[ii]==1)  /* it is an internal edge */
        {  
           InnerEdgeStat[kk]=ii+1; 
           kk++;   
        }

  
    /* the inner edges should not include those forcd edges, 
       thus revise  TotInnerEdgeNum here */
    TotInnerEdgeNum=kk;  
   

    kk=0;
    HybrdBoundEdgeNum=0;
    DielBoundEdgeNum=0;
    BoundEdgeStat = INT_Vector(TotBoundEdgeNum);


    /* find out boundary edges, those metal edges forming the contour
       of external conductors, which are charactered by MetalEdgeStat[i] 
       equaling to 3, are excluded  since J on this edges
       are equal to zero. */

    /* the following are dielectric boundary edges 
       EdgeStat[ii]==-1 means boundary edge 
       MetalEdgeStat[ii]==0 means it is not a metal patch, 
       instead it is a dielectric edge  */
    for(ii=0;ii<TotEdgeNum;ii++)
        if( (EdgeStat[ii]==-1) && (MetalEdgeStat[ii]==0) )
        { 
            DielBoundEdgeNum++;  
            HybrdBoundEdgeNum++;  
            BoundEdgeStat[kk]=ii+1; 
            kk++;
        }  /* if() */


    /* the following are half-metal/half-dielectric edges */
    for(ii=0;ii<TotEdgeNum;ii++)
        if( (EdgeStat[ii]==-1) && (MetalEdgeStat[ii]==1) )
        {   
            HybrdBoundEdgeNum++; 
            BoundEdgeStat[kk]=ii+1; 
            kk++; 
        }  /* if() */


    /* following are external metal edges */
    for(ii=0;ii<TotEdgeNum;ii++)
        if( (EdgeStat[ii]==-1) && (MetalEdgeStat[ii]==2) )
        {
            BoundEdgeStat[kk]=ii+1; 
            kk++;
        }  /* if() */

    /* this function can't be called before the boudary edge table
       is generated  */
    if (SourceType=='V')  DetectVSourceEdges();


    PrintOutput();
    return 0;

}   /* end of main() */



/**************************************************************************
Prototype:    void    ComputeParameters() 
Description:     To compute parameters for the mesh. For example, the 
                 total node  number, the total triangular element number, the 
                 total tetrahedron number. 
Input value:     none 
Return value:     none 
Global value used:     TotCubeNumEdge, TotCubeTrngleNum, TotCubeNumNode,
    TotNumHexHedron, TotTetElement, TotEdgeNum, TotNodeNum. TotTrngleNum,
    TotDipolNumEdge, TotDipolTrngleNum, Xdim, Ydim, Zdim, ExdcondNum, 
    Xextcond, Yextcond, DipolWid, DipolLen, Dipolplane_YZ, Dipolplane_XY,     
    TotInnerEdgeNum, InnerEdgeStat 
Global value modified:    TotCubeNumEdge, TotCubeTrngleNum, TotCubeNumNode,
    TotNumHexHedron, TotTetElement, TotEdgeNum, TotNodeNum. TotTrngleNum,
    TotInnerEdgeNum, InnerEdgeStat 
Subroutines called:     INT_Vector() 
***************************************************************************/
void ComputeParameters()
{

    int i;
    double error=1e-5;


    XdiM=YdiM=ZdiM=0;

    for(i=0; i<CellUnitXNum; i++)
        XdiM+=(int) ( (CellUnitX[i].end -CellUnitX[i].start)/CellUnitX[i].step +error );

    for(i=0; i<CellUnitYNum; i++)
        YdiM+=(int) ( (CellUnitY[i].end-CellUnitY[i].start)/CellUnitY[i].step +error );

    for(i=0; i<CellUnitZNum; i++)
        ZdiM+= (int) ( (CellUnitZ[i].end-CellUnitZ[i].start)/CellUnitZ[i].step +error );



    /* number of cubes */
    TotNumHexHedron = XdiM*YdiM*ZdiM;
    /* each cube can be devided into 5 tetrahedrons*/
    TotTetElement = 5 * TotNumHexHedron;

    /* number of edge in the dielectirc box */
    TotCubeNumEdge = 6*XdiM*YdiM*ZdiM + (XdiM+YdiM+ZdiM)
                    + 3*(XdiM*YdiM+YdiM*ZdiM+ZdiM*XdiM);

    /* number of triangles contained in the dielectirc box */
    TotCubeTrngleNum= 4*(XdiM*YdiM + YdiM*ZdiM + ZdiM*XdiM);
    TotCubeNumNode  = (XdiM+1)*(YdiM+1)*(ZdiM+1);
 
    /* TotEdgeNum, TotNodeNum and TotTrngleNum will also includes 
       those elements of external conductors */ 
    TotEdgeNum  = TotCubeNumEdge;
    TotNodeNum  = TotCubeNumNode;
    TotTrngleNum= TotCubeTrngleNum;


    /**************External Conductor information*************************/
    for(IEdge=0;IEdge<ExtcondNum;IEdge++)
    {
        /* This external conductor is a plane parallel to Y-Z plane */
        if( Xextcond[0][IEdge]==Xextcond[1][IEdge] )
        {
            DipolWid[IEdge]= (int) fabs((double)Yextcond[1][IEdge]-Yextcond[0][IEdge]) 
                            /DelYextcond[IEdge];
            DipolLen[IEdge]= (int)  fabs((double)Zextcond[1][IEdge]-Zextcond[0][IEdge])
                            /DelZextcond[IEdge];
            Dipolplane_YZ[IEdge]=1;

        }



        /* This external conductor is a plane parallel to X-Z plane */
        else if (Yextcond[0][IEdge]==Yextcond[1][IEdge])
        {
            DipolWid[IEdge]= (int) fabs((double)Xextcond[1][IEdge]-Xextcond[0][IEdge]) 
                             /DelXextcond[IEdge];
            DipolLen[IEdge]= (int) fabs((double)Zextcond[1][IEdge]-Zextcond[0][IEdge])
                             /DelZextcond[IEdge];
            Dipolplane_ZX[IEdge]=1;
        }


        /* This external conductor is a plane parallel to X-Y plane */
        else if (Zextcond[0][IEdge]==Zextcond[1][IEdge])
        {
            DipolWid[IEdge]= (int) (fabs((double)Xextcond[1][IEdge]-Xextcond[0][IEdge]) 
                             /DelXextcond[IEdge] );
            DipolLen[IEdge]= (int) (fabs((double)Yextcond[1][IEdge]-Yextcond[0][IEdge])
                             /DelYextcond[IEdge]);
            Dipolplane_XY[IEdge]=1;
        }

        else
        {
            fprintf(stderr, "Plate Orientation is not parallel to Axis.\n");  
            exit(1);
        }


        /* edges on the dipole */
        TotDipolNumEdge[IEdge] = 3*DipolLen[IEdge]*DipolWid[IEdge]
                       + DipolWid[IEdge] + DipolLen[IEdge];
                       
        /* Total edges equal to edges in the dielectrical box and
           edges on dipoles  */                      
        TotEdgeNum  = TotEdgeNum + TotDipolNumEdge[IEdge];

        /* total nodes on the dipole */
        TotDipolNumNode[IEdge] = (DipolLen[IEdge]+1)*(DipolWid[IEdge]+1);

        /* total nodes equal to nodes in the diectrical box and 
           nodes on the dipole */
        TotNodeNum  = TotNodeNum + TotDipolNumNode[IEdge];

        /* calculate total triangle number */
        TotDipolTrngleNum[IEdge]= 2*DipolLen[IEdge]*DipolWid[IEdge];
        TotTrngleNum= TotTrngleNum + TotDipolTrngleNum[IEdge]; 
 
    }  /* for(IEdge=0;IEdge<ExtcondNum;  */


    TotInnerEdgeNum = TotCubeNumEdge - 6*(XdiM*YdiM+YdiM*ZdiM+ZdiM*XdiM);
    InnerEdgeStat = INT_Vector(TotInnerEdgeNum);
 
}   /* end of ComputeParameter() */




/**************************************************************************
Prototype:    void    AllocateMemory() 
Description:     To dynamically allocate required memory. 
Input value:     none 
Return value:     none 
Global value used:     TotNumHexHedron, TotTetElement, TotTrngleNum,
    TotEdgeNum, TotNodeNum, MediaPerm, EdgeStat, MetalEdgeStat, EdgeFound,
    MultFound, PlusTrngleIndex, MinusTrngleIndex, GlobalEdgeEnds HexNode,
    HexEdgeNum, TetNode, TetEdge, TetBoundFaceTetHedronNum,
    TetBoundFaceFaceNum, TrngleNode, TrngleEdge, FaceNode, FaceEdge, EdgeStat,
    MetalEdgeStat, PlusTrngleDetect, PlusTrngleIndex, MinusTrngleDetect,
    MinusTrngleIndex, EdgeFound, MultFound ,GlobalEdgeEnds, NodeCord 
Global value modified:     HexNode, HexEdgeNum, TetNode, TetEdge,
    TetBoundFaceTetHedronNum, TetBoundFaceFaceNum, TrngleNode, TrngleEdge,
    FaceNode, FaceEdge, EdgeStat, MetalEdgeStat, PlusTrngleDetect,    
    PlusTrngleIndex, MinusTrngleDetect, MinusTrngleIndex, EdgeFound, 
    MultFound ,GlobalEdgeEnds, NodeCord 
Subroutines called:     INT_Matrix(), FLOAT_Matrix(), INT_Vector() 
**************************************************************************/
void AllocateMemory()
{

    int ii;

    HexNode = INT_Matrix(TotNumHexHedron,8);
    HexEdgeNum = INT_Matrix(TotNumHexHedron,19);
    MediaPerm   = double_Matrix(TotTetElement,2);
    TetNode  = INT_Matrix(TotTetElement,4);
    TetEdge  = INT_Matrix(TotTetElement,6);

    TetBoundFaceTetHedronNum = INT_Vector(TotTrngleNum);
    TetBoundFaceFaceNum = INT_Vector(TotTrngleNum);

    TrngleNode = INT_Matrix(TotTrngleNum,3);
    TrngleEdge = INT_Matrix(TotTrngleNum,3);

    FaceNode = INT_Matrix(TotTrngleNum,3);
    FaceEdge = INT_Matrix(TotTrngleNum,3);
    EdgeStat = INT_Vector(TotEdgeNum);
    MetalEdgeStat = INT_Vector(TotEdgeNum);
    PlusTrngleDetect = INT_Matrix(TotEdgeNum,3);
    PlusTrngleIndex  = INT_Vector(TotEdgeNum);
    MinusTrngleDetect = INT_Matrix(TotEdgeNum,3);
    MinusTrngleIndex  = INT_Vector(TotEdgeNum);
    EdgeFound  = INT_Vector(TotEdgeNum);
    MultFound  = INT_Vector(TotEdgeNum);
    GlobalEdgeEnds = INT_Matrix(TotEdgeNum,6);
    NodeCord = double_Matrix(TotNodeNum,3);

    /* the default permittivity is 1.0 + j0, which means in free space */
    for(ii=0;ii<TotTetElement;ii++) 
    {  
        MediaPerm[ii][0]=1.0;
        MediaPerm[ii][1]=0.0;
    }

    for(ii=0;ii<TotEdgeNum;ii++) 
    {
        EdgeStat[ii]=1; 
        MetalEdgeStat[ii]=0;
    }


    Null_INTVector( EdgeFound, TotEdgeNum );
    Null_INTVector( MultFound, TotEdgeNum );
    
    Null_INTVector( PlusTrngleIndex, TotEdgeNum );
    Null_INTVector( MinusTrngleIndex, TotEdgeNum );
    Null_INTMatrix( PlusTrngleDetect, TotEdgeNum, 3 );
    Null_INTMatrix( MinusTrngleDetect, TotEdgeNum, 3 );
    Null_INTMatrix( GlobalEdgeEnds, TotEdgeNum, 6 );

}



/**************************************************************************
Prototype:    void    AssignBoxCoordinates() 
Description:     To assign Cartesian coordinates for each node of 
                 each hexahedron 
Input value:     none 
Return value:     none 
Global value used:     NodeCord, Divisor 
Global value modified:     NodeCord 
Subroutines called: none 
**************************************************************************/
void AssignBoxCoordinates()
{
    int i, j, k, n, NodeNum;
    double Xtemp, Ytemp, Ztemp, DivisorX, DivisorY, DivisorZ;;

    
    NodeNum=0; 
     
    /* steps along three axises, revise here on Apr 10 */ 
    DivisorX = CellDimension*CellUnitX[0].step;
    DivisorY = CellDimension*CellUnitY[0].step;
    DivisorZ = CellDimension*CellUnitZ[0].step;
     
    /* start point */
    Xtemp    = CellDimension*XBoundDim[0]; 
    Ytemp    = CellDimension*YBoundDim[0]; 
    Ztemp    = CellDimension*ZBoundDim[0]; 
    
    for(k=0;k<=ZdiM;k++)
    {
        for(j=0;j<=YdiM;j++) 
        {
            for(i=0;i<=XdiM;i++) 
            {
                NodeCord[NodeNum][0] = Xtemp;
                NodeCord[NodeNum][1] = Ytemp;
                NodeCord[NodeNum][2] = Ztemp;
                NodeNum++;   
          
         
                /* to test whether needs to change step */
                for(n=0; n<CellUnitXNum; n++)
                    if(fabs((double)Xtemp-CellDimension*CellUnitX[n].start)<1e-5) break;
            
                /* change to new X step */  
                if(n!=CellUnitXNum) DivisorX=CellDimension*CellUnitX[n].step;
                    Xtemp = Xtemp + DivisorX; 
            }
        
            /* reset along x axis */
            Xtemp = CellDimension*XBoundDim[0];
            DivisorX = CellDimension*CellUnitX[0].step;
         
            /* increase 1 step along Y axis */
            /* revise here on Apr 10 */
            for(n=0; n<CellUnitYNum; n++)
                if(fabs((double)Ytemp-CellDimension*CellUnitY[n].start)<1e-5) break;
            
            /* change to new Y step */  
            if(n!=CellUnitYNum) DivisorY=CellDimension*CellUnitY[n].step;
                Ytemp = Ytemp + DivisorY; 
        
        }
       
        /* reset along y axis */
        Ytemp = CellDimension*XBoundDim[0]; 
        DivisorY = CellDimension*CellUnitY[0].step;
       
        /* increase 1 step along Z axis */
        for(n=0; n<CellUnitZNum; n++)
            if(fabs((double)Ztemp-CellDimension*CellUnitZ[n].start)<1e-5) break;
            
        /* change to new Z step */  
        if(n!=CellUnitZNum) DivisorZ=CellDimension*CellUnitZ[n].step;
          
        Ztemp = Ztemp + DivisorZ; 
    }
}




/**************************************************************************
Prototype:    void    AssignHexHedronNodeNum() 
Description:     To map the  local nodes to the  global nodes for each    
                 hexahedron. 
Input value:     none 
Return value:     none 
Global value used:     TotNumHexHedron, HexNode, XdiM, YdiM 
Global value modified:     none 
Subroutines called:     none **************************************************************************/
void AssignHexHedronNodeNum()
{
   int i, start;

   start=1;
  
   for(i=0;i<TotNumHexHedron;i++)
   {
      HexNode[i][0]=start;
      HexNode[i][1]=HexNode[i][0]+1;
      HexNode[i][2]=HexNode[i][0]+XdiM+1;
      HexNode[i][3]=HexNode[i][2]+1;
      HexNode[i][4]=HexNode[i][0]+(XdiM+1)*(YdiM+1);
      HexNode[i][5]=HexNode[i][4]+1;
      HexNode[i][6]=HexNode[i][4]+(XdiM+1);
      HexNode[i][7]=HexNode[i][6]+1;
      
      if( ((i+1)%((YdiM*XdiM))==0) ) start=start+(XdiM+3);
      else    start=start+1;

      if((start%(XdiM+1))==0) start++;

   }  /* for(i=0  */

}





/**************************************************************************
Prototype:    void    AssignHexHedronEdgeNum() 
Description:     To map the local edges to global edges for each hexahedron 
Input value:  none 
Return value:  none 
Global value used:  XdiM, YdiM, HexEdgeNum 
Global value modified:  none 
Subroutines called:  none 
**************************************************************************/
void AssignHexHedronEdgeNum()
{
    int EdgeNum,xx1,xx2,xx3,kk,dy1,dy2,divz,divz1;
  
    for(k=0;k<=ZdiM-1;k++) 
        for(j=0;j<=YdiM-1;j++)
            for(i=0;i<=XdiM-1;i++) 
            {  
                EdgeNum = k*YdiM*XdiM + j*XdiM + i;

                for(kk=0;kk<=17;kk++)   
                    HexEdgeNum[EdgeNum][kk] = 0;

                divz=(6*XdiM*YdiM+3*XdiM+3*YdiM+1)*k;
                divz1=  6*XdiM*YdiM + 3*XdiM + 3*YdiM + 1 ;
                dy1 = (3*XdiM+1)*j; dy2 = (3*XdiM+2)*j;

                xx1 = 1 + (YdiM+1)*XdiM + (2*XdiM+1)*YdiM;
                xx2 = xx1 + (2*XdiM+1);
                xx3 = xx2 + (XdiM + 1);

                HexEdgeNum[EdgeNum][0] = 1 + i + dy1 + divz;
                HexEdgeNum[EdgeNum][1] = 1 + XdiM + 2*i + dy1 + divz;
                HexEdgeNum[EdgeNum][2] = HexEdgeNum[EdgeNum][1] + 1;  
                HexEdgeNum[EdgeNum][3] = HexEdgeNum[EdgeNum][2] + 1;
                HexEdgeNum[EdgeNum][4] = 1 + 3*XdiM + 1 + i + dy1 + divz;
                HexEdgeNum[EdgeNum][5] = xx1 + 2*i + dy2 + divz;
                HexEdgeNum[EdgeNum][6] = HexEdgeNum[EdgeNum][5] + 1; 
                HexEdgeNum[EdgeNum][7] = HexEdgeNum[EdgeNum][6] + 1;
                HexEdgeNum[EdgeNum][8] = xx2 + i + dy2 + divz;
                HexEdgeNum[EdgeNum][9] = HexEdgeNum[EdgeNum][8] + 1;
                HexEdgeNum[EdgeNum][10]= xx3 + 2*i + dy2 + divz;
                HexEdgeNum[EdgeNum][11]= HexEdgeNum[EdgeNum][10] + 1;
                HexEdgeNum[EdgeNum][12]= HexEdgeNum[EdgeNum][11] + 1; 
                HexEdgeNum[EdgeNum][13]= HexEdgeNum[EdgeNum][0] + divz1;
                HexEdgeNum[EdgeNum][14]= HexEdgeNum[EdgeNum][1] + divz1; 
                HexEdgeNum[EdgeNum][15]= HexEdgeNum[EdgeNum][2] + divz1;
                HexEdgeNum[EdgeNum][16]= HexEdgeNum[EdgeNum][3] + divz1; 
                HexEdgeNum[EdgeNum][17]= HexEdgeNum[EdgeNum][4] + divz1;

            }  /* for(i=0 */
 
}




/**************************************************************************
Prototype:    void    DetectInnerEdgeStatus() 
Description:     To find edges located within the FEM region. 
Input value: none 
Return value: none 
Global value used: XdiM, YdiM, EdgeStat, HexEdgeNum, 
Global value modified: none 
Subroutines called: none ***************************************************************************/
void DetectInnerEdgeStatus()
{
 
     /* int EdgeNum,xx1,xx2,xx3,kk,dy1,dy2,divz,divz1; */
     int EdgeNum, i,j,k,m;
  
     for(k=0;k<=ZdiM-1;k++)
         for(j=0;j<=YdiM-1;j++)
             for(i=0;i<=XdiM-1;i++) 
             {       
       
                 EdgeNum = k*YdiM*XdiM + j*XdiM + i;
         
                 /* set default value to 1, which means it is an inner edge */
                 for(m=0; m<18; m++)  EdgeStat[HexEdgeNum[EdgeNum][m]-1] = 1;

                 /* detect whether it is a boundary edge */
                 if( (j==0)||(k==0) )  EdgeStat[HexEdgeNum[EdgeNum][0]-1] = -1;

                 if( (i==0)||(k==0) )  EdgeStat[HexEdgeNum[EdgeNum][1]-1] = -1;

                 if(k==0)  EdgeStat[HexEdgeNum[EdgeNum][2]-1] = -1;

                 if( (i==XdiM-1)||(k==0) )  
                     EdgeStat[HexEdgeNum[EdgeNum][3]-1] = -1;

                 if( (j==YdiM-1)||(k==0) )  
                     EdgeStat[HexEdgeNum[EdgeNum][4]-1] = -1;

                 if( (j==0)||(i==0) )   
                     EdgeStat[HexEdgeNum[EdgeNum][5]-1] = -1;

                 if( j==0 ) EdgeStat[HexEdgeNum[EdgeNum][6]-1] = -1;

                 if( (i==XdiM-1)||(j==0) )  
                     EdgeStat[HexEdgeNum[EdgeNum][7]-1] = -1;

                 if( i==0 ) EdgeStat[HexEdgeNum[EdgeNum][8]-1] = -1;

                 if( i==XdiM-1 )  EdgeStat[HexEdgeNum[EdgeNum][9]-1] = -1;

                 if( (j==YdiM-1)||(i==0) )  
                     EdgeStat[HexEdgeNum[EdgeNum][10]-1] = -1;

                 if( j==(YdiM-1) )   
                     EdgeStat[HexEdgeNum[EdgeNum][11]-1] = -1;

                 if( (i==XdiM-1)||(j==YdiM-1) )    
                       EdgeStat[HexEdgeNum[EdgeNum][12]-1] = -1;

                 if( (k==ZdiM-1)||(j==0) )   
                       EdgeStat[HexEdgeNum[EdgeNum][13]-1] = -1;

                 if( (k==ZdiM-1)||(i==0) )  
                       EdgeStat[HexEdgeNum[EdgeNum][14]-1] = -1;

                 if(k==ZdiM-1)  
                        EdgeStat[HexEdgeNum[EdgeNum][15]-1] = -1;

                 if((i==XdiM-1)||(k==ZdiM-1) )   
                        EdgeStat[HexEdgeNum[EdgeNum][16]-1] = -1;

                 if((j==YdiM-1)||(k==ZdiM-1) )  
                        EdgeStat[HexEdgeNum[EdgeNum][17]-1] = -1;
                             
            }  /* for() */
}
 
 

 
/**************************************************************************
Prototype:    void    DetectBoundEdgeStatus() 
Description:     To find edges located on the surface of the FEM region. 
Input value:     none 
Return value:     none 
Global value used:    XdiM, ZdiM, YdiM, EdgeStat, HexEdgeNum, BndcondNum, 
                      Xbndcond, Ybndcond, Zbndcond, MetalEdgeStat, EdgeStat,  
                      EdgeStat 
Global value modified: none
Subroutines called:     none
**************************************************************************/
void DetectBoundEdgeStatus()
{
    int EdgeNum;
    int  FrontWall, BackWall, BottomWall, TopWall, RightWall, LeftWall; 


    /*********************************************************
     assume the dielectric box is located in the orginal
     point the coordinate system is as follows:
  
                Y                    
                ^                     
                |                     
          X <--(x) Z                   
    *********************************************************/
  
    for(PatchNum=0;PatchNum<BndcondNum;PatchNum++)
    {
        FrontWall=0;BackWall=0;
        BottomWall=0;TopWall=0;
        RightWall=0;LeftWall=0;
 
        if( Zbndcond[0][PatchNum]==Zbndcond[1][PatchNum] )
        {
           if(Zbndcond[0][PatchNum]==ZBoundDim[0]) FrontWall=1;
           else   BackWall=1;
        }

        else if(Ybndcond[0][PatchNum]==Ybndcond[1][PatchNum])
        {
            if( Ybndcond[0][PatchNum]==YBoundDim[0] )  BottomWall=1;
            else    TopWall=1;
        }

        else if(Xbndcond[0][PatchNum]==Xbndcond[1][PatchNum]) 
        {
            if( Xbndcond[0][PatchNum]==XBoundDim[0] )  RightWall=1;
            else    LeftWall=1;
        }
         
        else 
        {
            fprintf(stderr,"Check conductor dimension\n");
            exit(1);    
        }
    
        /* revise here on Apr. 10 due to introducing non-uniform mesh */

        XmaxWall=YmaxWall= ZmaxWall=0;
        XminWall=YminWall= ZminWall=0;
   
        /* calculate XminWall */
        if( RightWall==1 )  XminWall=XmaxWall=0;
        else if( LeftWall==1 ) XminWall=XmaxWall=XdiM;
        else 
        {
            for(i=0; i<CellUnitXNum; i++)
                if( Xbndcond[0][PatchNum]>(CellUnitX[i].start-1e-5 ) &&
                    Xbndcond[0][PatchNum]<(CellUnitX[i].end +1e-5) )  break;

            if( i==CellUnitXNum )
            {
                fprintf(stderr,"Please check conductor's X dimension\n");
                exit(1);
            }
   
            for(j=0; j<i; j++)
                XminWall+= (int ) ( fabs((double)CellUnitX[j].end-CellUnitX[j].start) 
                           /CellUnitX[j].step+1e-5 );
            XminWall+= (int) ( (Xbndcond[0][PatchNum]-CellUnitX[i].start) 
                       /CellUnitX[i].step+1e-5 );
     
            /* calculate XmaxWall */
            for(i=0; i<CellUnitXNum; i++)
                if( Xbndcond[1][PatchNum]>(CellUnitX[i].start-1e-5) &&
                    Xbndcond[1][PatchNum]<(CellUnitX[i].end+1e-5 )) break;
            if( i==CellUnitXNum )
            {
                fprintf(stderr,"Check conductor's X dimension\n");
                exit(1);
            }
   
            for(j=0; j<i; j++)
                XmaxWall+= (int) ((CellUnitX[j].end-CellUnitX[j].start) 
                            /CellUnitX[j].step+1e-3 );
            XmaxWall += (int) ( (Xbndcond[1][PatchNum]-CellUnitX[i].start)
                        /CellUnitX[i].step+1e-3 );
        }/* else */
       
        if( BottomWall==1 )  YminWall=YmaxWall=0;
        else if(TopWall==1 ) YminWall=YmaxWall=YdiM;
        else 
        {    
            for(i=0; i<CellUnitYNum; i++)
                if( Ybndcond[0][PatchNum]>(CellUnitY[i].start-1e-5 ) &&
                    Ybndcond[0][PatchNum]<(CellUnitY[i].end +1e-5) )  break;
            if( i==CellUnitYNum )
            {
                fprintf(stderr,"Check conductor's Y dimension\n");
                exit(1);
            }
   
            for(j=0; j<i; j++)
                YminWall+= (int) ( fabs((double)CellUnitY[j].end-CellUnitY[j].start) 
                           /CellUnitY[j].step+1e-5 );

            YminWall += (int) ( (Ybndcond[0][PatchNum]-CellUnitY[i].start) 
                        /CellUnitY[i].step+1e-5 );
     
 
            /* calculate YmaxWall */
     
            for(i=0;i<CellUnitYNum; i++)
                if( Ybndcond[1][PatchNum]>(CellUnitY[i].start-1e-5) &&
                    Ybndcond[1][PatchNum]<(CellUnitY[i].end+1e-5 ) ) break;

            if( i==CellUnitYNum )
            {
                fprintf(stderr,"Check conductor's Y dimension\n");
                exit(1);
            }
   
            for(j=0; j<i; j++)
                YmaxWall+= (int ) ( fabs ((double)CellUnitY[j].end-CellUnitY[j].start) 
                        /CellUnitY[j].step+1e-5 );
            YmaxWall += (int) ( (Ybndcond[1][PatchNum]-CellUnitY[i].start) 
                        /CellUnitY[i].step+1e-5 );
     
        } /* else */
   
   
        /* determine ZminWall and ZmaxWall  */
        if( FrontWall==1 ) ZminWall=ZmaxWall=0;
        else if( BackWall==1 ) ZminWall=ZmaxWall=ZdiM;
        else
        {  
            for(i=0; i<CellUnitZNum; i++)
                if( Zbndcond[0][PatchNum]>(CellUnitZ[i].start-1e-5) &&
                    Zbndcond[0][PatchNum]<(CellUnitZ[i].end +1e-5) ) break;
            if( i==CellUnitZNum )
            {
                fprintf(stderr,"Check conductor's Z dimension\n");
                exit(1);
            }
   
            for(j=0; j<i; j++)
                ZminWall+= (int ) ( fabs((double)CellUnitZ[j].end-CellUnitZ[j].start) 
                          /CellUnitZ[j].step+1e-5 );
            ZminWall += (int ) ( (Zbndcond[0][PatchNum]-CellUnitZ[i].start) 
                          /CellUnitZ[i].step+1e-5 );
     
 
            /* calculate ZmaxWall */
            for(i=0; i<CellUnitZNum; i++)
                if( Zbndcond[1][PatchNum]>(CellUnitZ[i].start-1e-5) &&
                    Zbndcond[1][PatchNum]<(CellUnitZ[i].end+1e-5) ) break;

            if( i==CellUnitZNum )
            {
                fprintf(stderr,"Please check conductor's Z dimension\n");
                exit(1);
            }
   
            for(j=0; j<i; j++)
                ZmaxWall+= (int ) ( fabs((double)CellUnitZ[j].end-CellUnitZ[j].start)/
                         CellUnitZ[j].step+1e-3 );

            ZmaxWall += (int ) ( (Zbndcond[1][PatchNum]-CellUnitZ[i].start)/ 
                         CellUnitZ[i].step+1e-3 );

        } /* end of else */    


        if( BottomWall )
        {
 
            j=YminWall;
            for(k=0;k<ZdiM;k++)
                for(i=0;i<XdiM;i++) 
                {

                    EdgeNum = k*YdiM*XdiM + j*XdiM + i;
                    if( (i>=XminWall) && (i<XmaxWall) && 
                        (k>=ZminWall) && (k<ZmaxWall) )
                    {
                        MetalEdgeStat[HexEdgeNum[EdgeNum][0]-1]++;
                        MetalEdgeStat[HexEdgeNum[EdgeNum][5]-1]++;
                        /* diagonal edge is always external metal edge */
                        MetalEdgeStat[HexEdgeNum[EdgeNum][6]-1] =2;          
                        MetalEdgeStat[HexEdgeNum[EdgeNum][7]-1]++;
                        MetalEdgeStat[HexEdgeNum[EdgeNum][13]-1]++;
                    }
                }

        }

        if( TopWall )
        {
            j=YmaxWall-1;
            for(k=0;k<ZdiM;k++)
                for(i=0;i<XdiM;i++) 
                {
                    EdgeNum = k*YdiM*XdiM + j*XdiM + i;
                    if( (i>=XminWall) && (i<XmaxWall) &&  
                        (k>=ZminWall) && (k<ZmaxWall) )
                    {

                        MetalEdgeStat[HexEdgeNum[EdgeNum][4]-1] ++;
                        MetalEdgeStat[HexEdgeNum[EdgeNum][10]-1]++;

                        /* diagonal edge */
                        MetalEdgeStat[HexEdgeNum[EdgeNum][11]-1]=2;            
                        MetalEdgeStat[HexEdgeNum[EdgeNum][12]-1]++;
                        MetalEdgeStat[HexEdgeNum[EdgeNum][17]-1]++;
                    }  /* if( (!>=XminWall */
                }      /* for(i=0; */
 
        }  /* if(TotWall */

        if( FrontWall )
        {
            k=ZminWall;
            for(j=0;j<YdiM;j++)
                for(i=0;i<XdiM;i++) 
                {
                    EdgeNum = k*YdiM*XdiM + j*XdiM + i;
                    if( (i>=XminWall) && (i<XmaxWall) &&
                        (j>=YminWall) && (j<YmaxWall) )
                    {

                        MetalEdgeStat[HexEdgeNum[EdgeNum][0]-1] ++;
                        MetalEdgeStat[HexEdgeNum[EdgeNum][1]-1] ++;
                        /* diagonal edge */
                        MetalEdgeStat[HexEdgeNum[EdgeNum][2]-1] =2;     
                        MetalEdgeStat[HexEdgeNum[EdgeNum][3]-1] ++;
                        MetalEdgeStat[HexEdgeNum[EdgeNum][4]-1] ++;
                    }  /* if( (i>=XminWall */
                }      /* for(i=0 */
 
        }

        if( BackWall )
        {
 
            k=ZmaxWall-1;
            for(j=0;j<YdiM;j++)
                for(i=0;i<XdiM;i++) 
                {
                    EdgeNum = k*YdiM*XdiM + j*XdiM + i;
                    if( (i>=XminWall) && (i<XmaxWall) && 
                        (j>=YminWall) && (j<YmaxWall))
                    {

                        MetalEdgeStat[HexEdgeNum[EdgeNum][13]-1] ++;
                        MetalEdgeStat[HexEdgeNum[EdgeNum][14]-1] ++; 
                        /* diagonal edge */
                        MetalEdgeStat[HexEdgeNum[EdgeNum][15]-1] =2;          
                        MetalEdgeStat[HexEdgeNum[EdgeNum][16]-1] ++;
                        MetalEdgeStat[HexEdgeNum[EdgeNum][17]-1] ++;
                    }
                }

        }

        if( RightWall )
        {
            i=XminWall;
            for(k=0;k<ZdiM;k++)
                for(j=0;j<YdiM;j++)
                {
                    EdgeNum = k*YdiM*XdiM + j*XdiM + i;
                    if( (j>=YminWall) && (j<YmaxWall) &&  
                        (k>=ZminWall ) && (k<ZmaxWall) )
                    {

 	                MetalEdgeStat[HexEdgeNum[EdgeNum][1]-1]++;
 	                MetalEdgeStat[HexEdgeNum[EdgeNum][5]-1]++;
 	                MetalEdgeStat[HexEdgeNum[EdgeNum][8]-1] =2;
 	                MetalEdgeStat[HexEdgeNum[EdgeNum][10]-1]++;
 	                MetalEdgeStat[HexEdgeNum[EdgeNum][14]-1]++;
                    }
                }
 
        }

        if( LeftWall )
        {
            i=XmaxWall-1;
            for(k=0;k<ZdiM;k++)
                for(j=0;j<YdiM;j++)
                {
                    EdgeNum = k*YdiM*XdiM + j*XdiM + i;
                    if( (j>=YminWall) && (j<YmaxWall) &&  
                        (k>=ZminWall) && (k<ZmaxWall) )
                    {
                        MetalEdgeStat[HexEdgeNum[EdgeNum][3]-1] ++;
                        MetalEdgeStat[HexEdgeNum[EdgeNum][7]-1] ++; 
                        /* diagonal edge */
                        MetalEdgeStat[HexEdgeNum[EdgeNum][9]-1] =2; 
                        MetalEdgeStat[HexEdgeNum[EdgeNum][12]-1]++;
                        MetalEdgeStat[HexEdgeNum[EdgeNum][16]-1]++;
                    }  /* if( (j>=YminWall */
                }    /* for */
  
        } /* if(LeftWall) */

    } /* for( PatchNum */

  
}   /* end of   DetectBoundEdgeStatus()  */




/***************************************************************************
Prototype:  void  Tetra_Edge_Div1() 
Description:  To divide one hexahedron into five tetrahedra. There are two  
              methods to do that. This subroutine uses one method.   
              Tetra_Edge_Div2() use the other one. 
Input value:   none 
Return value:   none 
Global value used:   HexNum, TetNode, HexNode, TetEdge, HexEdgeNum.
Global value modified:    none 
Subroutines called:     none 
***************************************************************************/
void TetraEdgeDiv1()
{
    int Count, i, j;

    /* map a hexhedron into five tetrahedron */
    int TetraNode[][4]={ {0, 3, 2, 6}, {3, 5, 7, 6}, {0, 1, 3, 5},
                         {0, 6, 4, 5}, {0, 5, 3, 6} };

    /* generate the tetrahedron edge table from the hexhedron table */

    /* the sign of the tetrahedron edges */
    int TetraEdgeSign[][6]={ {1, 1, 1, -1, -1, 1}, {1, 1, 1,  1, -1, -1},
                             {1, 1, 1,  1, -1, 1}, {1, 1, 1, -1,  1,  1},
                             {1, 1, 1, -1, -1, 1} };

    /* the map table among tetrahedra and the hexhedron */
    int TetraEdge[][6]={ {2, 1, 8, 4, 11, 10}, {9, 12, 11, 16, 15, 17},
                         {0, 2, 6, 3, 7,  9},  {8, 5, 6, 14, 15, 13},
                         {6, 2, 8, 9, 15, 11}  };


    Count = 5*HexNum;  /* one hexhedron is divided into five tetrahedra */

    
    for(i=0; i<5; i++) 
    {  
        /* the ith tetrahedron */

        for(j=0; j<4; j++) /* counter of the node number of one tetrahedron */
            TetNode[Count+i][j]=HexNode[HexNum][ TetraNode[i][j] ];

        for(j=0; j<6; j++) /* counter of the edge number of one tetrahedron */
            TetEdge[Count+i][j] = TetraEdgeSign[i][j]*
                HexEdgeNum[HexNum][ TetraEdge[i][j] ];

    }  /* for(i=0 */

}



/****************************************************************************
Prototype:    void    Global_Edge_Ends_Div1() 
Description:  To assign start and end nodes for the 18 edges associated with one
              hexahedron. This subroutine is used accompanying with  
              Tetra-Edge_Div1(). 
Input value:     none 
Return value:     none 
Global value used:     GlobalEdgeEnds, HexNode, HexNum, 
Global value modified:     none 
Subroutines called:     none 
****************************************************************************/
void GlobalEdgeEndsDiv1()
{
    int i;  /* counter */
    
    int StartNode[]={0, 0, 0, 1, 2, 0, 0, 1, 0, 3, 2, 3, 3, 4, 4, 5, 5, 6 },
        EndNode[]=  {1, 2, 3, 3, 3, 4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 7 };
 
    for(i=0; i<18; i++) 
    {
        int EdgeNum = HexEdgeNum[HexNum][i];  /* the edge number */
        /* find out the start node */
        GlobalEdgeEnds[EdgeNum-1][0]=HexNode[HexNum][ StartNode[i] ];
        /* findout the end node */
        GlobalEdgeEnds[EdgeNum-1][1]=HexNode[HexNum][ EndNode[i] ];
    }
}




/****************************************************************************
Prototype:  void  Tetra_Edge_Div2() 
Description:  To divide one hexahedron into five tetrahedron. There are two  
              methods to do that. Tetra_Edge_Div1() uses 
              one method, Tetra_Edge_Div2() uses the other method. 
Input value:  none 
Return value:  none 
Global value used:   HexNum, TetNode, HexNode, TetEdge, HexEdgeNum, 
Global value modified:   none 
Subroutines called:  none *****************************************************************************/
void TetraEdgeDiv2()
{
    int Count;

    /* map a hexhedron into five tetrahedron */
    int TetraNode[][4]={ {0, 1, 2, 4}, {1, 7, 4, 5}, {1, 7, 3, 2},
                         {2, 6, 4, 7}, {1, 2, 4, 7} };

    /* generate the tetrahedron edge table from the hexhedron table */

    /* the sign of the tetrahedron edges */
    int TetraEdgeSign[][6]={ {1, 1, 1, 1, -1, 1}, {1, 1, 1,  -1, 1, 1},
                             {1, 1, 1,  -1, 1, -1}, {1, 1, 1, -1,  -1,  1},
                             {1, 1, 1, 1, -1, 1} };

    /* the map table among tetrahedra and the hexhedron */
    int TetraEdge[][6]={ {0, 1, 5, 2, 6, 8}, { 9, 6, 7, 15, 16, 13},
                         {9, 3, 2, 12, 11, 4}, {10, 8, 11, 14, 17, 15}, 
                         {2, 6, 9, 8, 11, 15}  };

    Count = 5*HexNum;

    for(i=0; i<5; i++) 
    {  
        /* the ith tetrahedron */

        for(j=0; j<4; j++) /* counter of the node number of one tetrahedron */
            TetNode[Count+i][j]=HexNode[HexNum][ TetraNode[i][j] ];

        for(j=0; j<6; j++) /* counter of the edge number of one tetrahedron */
            TetEdge[Count+i][j] = TetraEdgeSign[i][j]*
                HexEdgeNum[HexNum][ TetraEdge[i][j] ];

    }  /* for(i=0 */


}   /* end of TetraEdgeDiv2()  */



/***************************************************************************
Prototype:    void    Global_Edge_Ends_Div2() 
Description:  To assign start and end nodes for the 18 edges associated 
              with one hexahedron. This subroutine is 
              used accompanying with Tetra-Edge_Div2(). 
Input value:     none 
Return value:     none 
Global value used:     GlobalEdgeEnds, HexNode, HexNum, 
Global value modified:     none 
Subroutines called:     none
***************************************************************************/
void GlobalEdgeEndsDiv2()
{

    int i;  /* counter */
    
    int StartNode[]={0, 0, 1, 1, 2, 0, 1, 1, 2, 1, 2, 2, 3, 4, 4, 4, 5, 6 },
        EndNode[]=  {1, 2, 2, 3, 3, 4, 4, 5, 4, 7, 6, 7, 7, 5, 6, 7, 7, 7 };
 
    for(i=0; i<18; i++) 
    {
        int EdgeNum = HexEdgeNum[HexNum][i];  /* the edge number */
        /* find out the start node */
        GlobalEdgeEnds[EdgeNum-1][0]=HexNode[HexNum][ StartNode[i] ];
        /* findout the end node */
        GlobalEdgeEnds[EdgeNum-1][1]=HexNode[HexNum][ EndNode[i] ];
    }  /* for(i=0 */

}  /* end of GlobalEdgeEndDiv2() */





/**************************************************************************
Prototype:    void    FindTetraHedronFaceNodes(int TetHedronNum) 
Description:   To find tetrahedron faces that lie on the surface of FEM   
               region. 
Input value:     
    int TetHedronNum--- index of the tetrahedron 
Return value:     none 
Global value used:     EdgeStat,TetNode, NodeCord 
Global value modified:     none 
Subroutines called:     none
***************************************************************************/
void FindTetraHedronFaceNodes( int TetHedronNum)
{
    int Count_i,Count_j,NCount_j,NCount_k,NCount_l;
    int Duplicate=0,ECount_j,ECount_k,ECount_l;
    int TrngleEdge0,TrngleEdge1,TrngleEdge2;
    int TrngleNode0,TrngleNode1,TrngleNode2;
    int TrngleNodeX,TrngleNodeY,TrngleNodeZ;

    for(Count_i=0;Count_i<=3;Count_i++)
    {
       
        if( Count_i==0 ) 
        {
            ECount_j=0; 
            ECount_k=1; 
            ECount_l=3;    
            NCount_j=2; 
            NCount_k=1; 
            NCount_l=0;

        }
        else if(Count_i==1) 
        {   
            ECount_j=0;
            ECount_k=4; 
            ECount_l=2; 
            NCount_j=3; 
            NCount_k=0; 
            NCount_l=1;

        }
        else if(Count_i==2) 
        {
            ECount_j=3; 
            ECount_k=5;
            ECount_l=4; 
            NCount_j=3; 
            NCount_k=1; 
            NCount_l=2;

        }
        else                
        {
            ECount_j=1; 
            ECount_k=2; 
            ECount_l=5;
            NCount_j=3; 
            NCount_k=2; 
            NCount_l=0;
        }
      
        TrngleEdge0 = EdgeStat[abs(TetEdge[TetHedronNum][ECount_j])-1];
        TrngleEdge1 = EdgeStat[abs(TetEdge[TetHedronNum][ECount_k])-1];
        TrngleEdge2 = EdgeStat[abs(TetEdge[TetHedronNum][ECount_l])-1];

        TrngleNode0 = TetNode[TetHedronNum][NCount_j]-1;
        TrngleNode1 = TetNode[TetHedronNum][NCount_k]-1;
        TrngleNode2 = TetNode[TetHedronNum][NCount_l]-1;

        if((((fabs((double)NodeCord[TrngleNode0][0]-NodeCord[TrngleNode1][0])<1e-5)&&
             (fabs((double)NodeCord[TrngleNode0][0]-NodeCord[TrngleNode2][0])<1e-5))
         ||((fabs((double)NodeCord[TrngleNode0][1]-NodeCord[TrngleNode1][1])<1e-5) &&
            (fabs((double)NodeCord[TrngleNode0][1]-NodeCord[TrngleNode2][1])<1e-5) )
         ||((fabs((double)NodeCord[TrngleNode0][2]-NodeCord[TrngleNode1][2])<1e-5) &&
            (fabs((double)NodeCord[TrngleNode0][2]-NodeCord[TrngleNode2][2])<1e-5) )
          )
          && ((TrngleEdge0==-1) && (TrngleEdge1==-1) && (TrngleEdge2==-1))
        )
        {
            Duplicate = 0;

            for(Count_j=0;Count_j<=TotTetFaceNum;Count_j++)
            {
                TrngleNodeX = TrngleNode[Count_j][0]-1;
                TrngleNodeY = TrngleNode[Count_j][1]-1;
                TrngleNodeZ = TrngleNode[Count_j][2]-1;
                if( ( (TrngleNode0==TrngleNodeX) && 
                    (((TrngleNode1==TrngleNodeY) && 
                      (TrngleNode2==TrngleNodeZ)) 
                   || ((TrngleNode1==TrngleNodeZ) &&  
                      (TrngleNode2==TrngleNodeY)))) 
                   || ((TrngleNode1==TrngleNodeY) && 
                     (((TrngleNode0==TrngleNodeX) && 
                       (TrngleNode2==TrngleNodeZ)) 
                   || ((TrngleNode0==TrngleNodeZ) && 
                       (TrngleNode2==TrngleNodeX))))
                   ||( (TrngleNode2==TrngleNodeZ) && 
                     (((TrngleNode0==TrngleNodeX) && 
                       (TrngleNode1==TrngleNodeY)) 
                   || ((TrngleNode0==TrngleNodeY) && 
                       (TrngleNode1==TrngleNodeX)))
                    )
                  )
                {
                     Duplicate = Count_j + 1; 
                     break;
                }   /* if ()  */

                else 
                {   
                    Duplicate = 0; 
                    continue;
                }  /* else */
            }      /* for() */ 

            if( Duplicate==0 )
            {
                int i=TotTetFaceNum, j=TotTetFaceNum, k=TetHedronNum;
 
                TrngleNode[i][0]=TetNode[k][NCount_j];
                TrngleNode[i][1]=TetNode[k][NCount_k];
                TrngleNode[i][2]=TetNode[k][NCount_l];
                TrngleEdge[j][0]=TetEdge[k][ECount_j];
                TrngleEdge[j][1]=TetEdge[k][ECount_k];
                TrngleEdge[j][2]=TetEdge[k][ECount_l];
                TetBoundFaceTetHedronNum[j]=k+1;
                TetBoundFaceFaceNum[j]=Count_i+1;
                TotTetFaceNum++;

            }   /* if( Duplicate==0 ) */
        }       /* if( ) */      
    }           /* for(Count */

}               /* end of FindTetrahedronFaceNodes() */    





/*****************************************************************************
Prototype:    void    ModifyBoundFaces(int TrngleFace) 
Description:    To orient the edges of each tetrahedron face to consistent 
                with   the reference direction 
Input value: 
    int TrngleNum--- the index of the boundary face 
Return value: none 
Global value used:     TrngleNode, GlobalEdgeEnds, 
Global value modified:     none 
Subroutines called:     none 
******************************************************************************/
void ModifyBoundFaces(int TrngleFace)
{
    int i,Node1,Node2,Node3,Node4,Edge1;

    for(i=0;i<=2;i++)
    {
        Node1= TrngleNode[TrngleFace][i];
        Node2= TrngleNode[TrngleFace][(i+1)%3];
        Edge1= abs(TrngleEdge[TrngleFace][(i+2)%3]);
        Node3= GlobalEdgeEnds[Edge1-1][0];
        Node4= GlobalEdgeEnds[Edge1-1][1];

        if( (Node1==Node3) && (Node2==Node4) )
            TrngleEdge[TrngleFace][(i+2)%3] = Edge1;
        else TrngleEdge[TrngleFace][(i+2)%3] = -Edge1;

    }  /* for(i=0 */
}      /* end if ModifyBoundFaces()  */
 
 


/***************************************************************************
Prototype:    void    ComputeTrngleCentroid(int, double **, int **, double **) 
Description:    To compute the centroid of a triangular patch 
Input value: 
    int TrngleNum  --- the triangle number 
    double **Cord,  --- the global node table 
    int **BFaceNode --- the node of the triangle 
    double **CenTroid --- where the results are stored. 
Return value: none 
Global value used: none 
Global value modified: none 
Subroutines called: VTXadd2() 
*****************************************************************************/
void ComputeTrngleCentroid(int TrngleNum,double **Cord, int **BFaceNode,
                           double **CenTroid)
{
    int Count_k; double Buff[3],Buff1[3],Buff2[3],Buff3[3];
 
    for(Count_k=0;Count_k<=2;Count_k++)
    {
        Buff1[Count_k]=Cord[BFaceNode[TrngleNum][0]-1][Count_k];
        Buff2[Count_k]=Cord[BFaceNode[TrngleNum][1]-1][Count_k];
        Buff3[Count_k]=Cord[BFaceNode[TrngleNum][2]-1][Count_k];

    }  /* for(Count_k  */

    VTXadd2(Buff1,Buff2,Buff3,Buff);
    CenTroid[TrngleNum][0] = (1.0/3.0) * Buff[0];
    CenTroid[TrngleNum][1] = (1.0/3.0) * Buff[1];
    CenTroid[TrngleNum][2] = (1.0/3.0) * Buff[2];

}   /* end of ComputeTrngleCentroid() */
 
 
 


/*************************************************************************
Prototype:    void    Read_SIF_File() 
Description:    Reads input file 
Input value:     none 
Return value:     none 
Global value used: CellDim, XdiM, YdiM, ZdiM, XDielmat, YDielmat, 
                   ZDielmat,  Xintcond, Yintcond, Zintcond, BndcondNum,  
                   Xextcond, Yextcond, Zextcond, XextVsrc, YextVsrc, 
                   ZextVsrc, XintEsrc, YintEsrc, ZintEsrc, ExtcondNum, 
                   BndcondNum, ExtVsrcNum, freq, SourceType, PolInc, 
                   ThetaInc, PhiInc, OperateFreq, CellDimension , Min_X, 
                   Min_Y, Min_Z, Epsilon, Sigma, DielmatNum, IntcondNum, 
                   VsourceMag, IntEsrcNum. 
Global value modified: Min_X, Min_Y, Min_Z, XdiM, YdiM, ZdiM, CellDim,  
                       DielmatNum, IntcondNum, BndcondNum, ExtcondNum,  
                       ExtVsrcNum, VsourceMag, IntEsrcNum, ThetaInc,
                       PhiInc, OperateFreq, CellDimension. 
Subroutines called: none 
*************************************************************************/
void ReadSIFFile()
{
    int    x1, y1, z1, x2, y2, z2, delx, dely, delz, i;
    char   axis, type[20], buffer[80];
    double  tmp1, tmp2, freq = 0, CellDim = 1;

    Min_X = 10000; Min_Y=10000; Min_Z=10000;
    Max_X = 0;     Max_Y=0;     Max_Z=0;
    CondNum = 0;
 
    printf("\nProcessing Key Words:\n\n");
 
    while ( fscanf(InF,"%s", type)!=EOF )
    {
        /* # line means comments */
        if( type[0] == '#' ) 
        {
            /* get rid of those comment lines */
            fgets(buffer, 80, InF);
            continue;
        }

        if( !strncmp(type, "unit", 4) ) 
        {
             printf("\tunit\n");
 
            if( fscanf(InF, "%lf %s", &tmp1, buffer)!=2 )
            {
                fprintf(stderr, "unit: missing some parameters\n");
                exit(1);
            }

            if( !strcmp(buffer, "cm") ) tmp2=0.01;
            else if(!strcmp(buffer, "m"))  tmp2=1;
            else if(!strcmp(buffer, "mm")) tmp2=0.001;
            else 
            {
                fprintf(stderr,"ERROR: unrecognized cell dimension.\n");
                exit(1);
            }
            CellDim=tmp1*tmp2;  /* get the unit in meter */
            continue; 
        }   

        if( !strcmp(type,"boundary") ) 
        {
            printf("\tboundary\n");
            if( fscanf(InF, "%d %d %d %d %d %d", &x1, &y1, &z1,
                &x2, &y2, &z2)!=6 )
            {
                fprintf(stderr,"Boundary: missing some parameters\n");
                exit(1);
            }      
  
            XBoundDim[0] = x1; 
            XBoundDim[1] = x2;
            YBoundDim[0] = y1; 
            YBoundDim[1] = y2;
            ZBoundDim[0] = z1; 
            ZBoundDim[1] = z2;
        
	    /* to check the boundary parameters */
            if( (x1==x2)||(y1==y2)||(z1==z2) ) 
            {
                fprintf(stderr, "boundary: dimension is invalid.\n");
                exit(1);
            }
            continue;
        
        } /* end of "boundary" */
        
        if( !strcmp(type,"conductor") )
        {
            printf("\tconductor\n");
            if( fscanf(InF, "%d %d %d %d %d %d %d %d %d", &x1, &y1, &z1,
                          &x2, &y2, &z2, &delx, &dely, &delz)!=9 )
            {
                fprintf(stderr, "conductor: missing some parameters\n");
                exit(1);
            }

            /* check the parameters */
            if( (x1!=x2) && (y1!=y2) &&(z1!=z2) )
            {
           	fprintf(stderr, "Conductor should be two demension and \
parallel one axis!\n");
           	exit(1);
            }
           
            conductor[CondNum].x1=x1;
            conductor[CondNum].y1=y1;
            conductor[CondNum].z1=z1;
            conductor[CondNum].x2=x2;
            conductor[CondNum].y2=y2;
            conductor[CondNum].z2=z2;
            conductor[CondNum].DelX= delx;
            conductor[CondNum].DelY= dely;
            conductor[CondNum].DelZ= delz;
            
            CondNum++;
            continue;
        }   /* end of "conductor */    
              
       
        if( !strcmp(type, "dielectric") ) 
        {
            printf("\tdielectric\n");
            if( fscanf(InF, "%d %d %d %d %d %d %lf %lf", &x1, &y1, &z1, 
                    &x2, &y2, &z2, &tmp1, &tmp2)!=8 )
            {
                fprintf(stderr,"dielectric: missing some parameters\n");
                exit(1);
            }
            if( (x1==x2)||(y1==y2)||(z1==z2) )
            {
                fprintf(stderr,"ERROR:  diel thickness is not finite.\n");
                exit(1);
            }
            XDielmat[0][DielmatNum]=x1;
            XDielmat[1][DielmatNum]=x2;
            YDielmat[0][DielmatNum]=y1;
            YDielmat[1][DielmatNum]=y2;
            ZDielmat[0][DielmatNum]=z1;
            ZDielmat[1][DielmatNum]=z2;
            Epsilon[DielmatNum]=tmp1; 
            Sigma[DielmatNum]=tmp2;
            DielmatNum++;
            continue;
        }  /* end "dielectric" */

        if( !strcmp(type,"vsource") )
        {
            printf("\tvsource\n");
            SourceType='V';
            if( fscanf(InF, "%d %d %d %d %d %d %lf %c %lf",
                &x1, &y1, &z1, &x2, &y2, &z2, &freq, &axis, &tmp1)!=9 )
            {
                fprintf(stderr, "vsource: missing some parameters\n");
                exit(1);
            }

            if( (x1==x2)&&(y1==y2)&&(z1==z2) ) 
            {
               fprintf(stderr, "ERROR: vsource cannot be defined at \
a point.\n"); 
               exit(1);
            }
            if( (axis != 'x')  && (axis!= 'y') && (axis != 'z') )
            {
                fprintf(stderr, "vsource: unrecognized polarization\n");
                exit(1);
            }
 
            XextVsrc[0][ExtVsrcNum] = x1;
            XextVsrc[1][ExtVsrcNum] = x2;
            YextVsrc[0][ExtVsrcNum] = y1;
            YextVsrc[1][ExtVsrcNum] = y2;
            ZextVsrc[0][ExtVsrcNum] = z1;
            ZextVsrc[1][ExtVsrcNum] = z2;
            ExtVsrcMag[ExtVsrcNum] = tmp1;
            ExtVsrcNum++;
            
            continue;
        }   /* end "vsource" */
   
        if( !strcmp(type,"isource") ) 
        {
            SourceType='I';
            printf("\tisource\n");
            if( fscanf(InF, "%d %d %d %d %d %d  %lf %c %lf",
                &x1, &y1, &z1, &x2, &y2, &z2, &freq, &axis,&tmp1)!=9)
            {
                fprintf(stderr, "isource: missing some parameters\n");
                exit(1);
            }

            XintEsrc[0][IntEsrcNum] = x1;
            XintEsrc[1][IntEsrcNum] = x2;
            YintEsrc[0][IntEsrcNum] = y1;
            YintEsrc[1][IntEsrcNum] = y2;
            ZintEsrc[0][IntEsrcNum] = z1; 
            ZintEsrc[1][IntEsrcNum] = z2;
            IntEsrcAxis[IntEsrcNum]=axis;
            IntEsrceMag[IntEsrcNum]=tmp1;        
            IntEsrcNum++;
        
            if( (x1==x2)&&(y1==y2)&&(z1==z2) ) 
            {
        	fprintf(stderr, "ERROR: isource cannot be defined \
at a point.\n"); 
                exit(1);
            }

            if( (axis != 'x') && (axis!= 'y') && ( axis != 'z') )
            {
        	fprintf(stderr, "isource: unrecognized polarization\n");
                exit(1); 
            }
            continue;

        }  /* end "esource" */

        if( !strcmp(type,"eplane") )
        {
            printf("\teplane\n");
            if( fscanf(InF, "%lf %lf %lf  %lf %lf %lf", &freq, &E_Theta, \
&E_Phi,  &K_Theta, &K_Phi, &PlaneWaveE_Mag)!=6 )
            {
                fprintf(stderr,"%lf  %lf %lf  %lf %lf %lf", freq, E_Theta, \
E_Phi, K_Theta, K_Phi, PlaneWaveE_Mag );
                fprintf(stderr, "eplane: missing some parameters\n");
                exit(1);
            }

            SourceType='P';
            continue;
        } /* end "eplane" */
    
    
        if( !strcmp(type,"celldim") )
        {
            char axis;
            double start, end, step;
      
            printf("\tcelldim\n");
      
            if( fscanf(InF, "\t %lf  %lf %lf %c", &start, &end, &step, \
&axis)!=4 )
            {
                fprintf(stderr, "celldim: missing some parameters\n");
                exit(1);
            }
      
            if( axis=='x' )
            {
               CellUnitX[CellUnitXNum].start=start;
               CellUnitX[CellUnitXNum].end=end;
               CellUnitX[CellUnitXNum].step=step;
               CellUnitXNum++;
            }
            else if(axis=='y')
            {
                CellUnitY[CellUnitYNum].start=start;
                CellUnitY[CellUnitYNum].end=end;
                CellUnitY[CellUnitYNum].step=step;
                CellUnitYNum++;
            }
      
            else if( axis=='z' )
            {
                CellUnitZ[CellUnitZNum].start=start;
                CellUnitZ[CellUnitZNum].end=end;
                CellUnitZ[CellUnitZNum].step=step;
                CellUnitZNum++;
            }
        
            else 
            {
                fprintf(stderr,"celldim: unrecognized axix\n");
                exit(1);
            }
            continue;
      
        }  

        if( !strcmp(type, "resistor") ) 
        {
            double tmp;

            printf("\tresistor\n");
            fscanf(InF,"%d %d %d %d %d %d %lf", &x1, &y1, &z1, &x2, &y2, \
&z2, &tmp);
            ResisterLocX[ResisterNum][0]=x1;
            ResisterLocX[ResisterNum][1]=x1;
            ResisterLocY[ResisterNum][0]=y1;
            ResisterLocY[ResisterNum][1]=y2;
            ResisterLocZ[ResisterNum][0]=z1;
            ResisterLocZ[ResisterNum][1]=z2;
            ResisterValue[ResisterNum]=tmp;
            ResisterNum++;
            continue;
        }

        if( !strcmp(type,"output") ) 
        {
   	    char filename[20];
   	    char axis;
   	
   	    printf("\toutput\n");
            if( fscanf(InF, "%d %d %d %d %d %d  %c %s", &x1, &y1, &z1, 
                      &x2, &y2, &z2, &axis, filename)!=8 )
            {
                fprintf(stderr, "output: misisng some parameters\n");
                exit(1);
            }
        
            if( axis!='x' && axis!='y'&& axis!='z' ) 
            {
                fprintf(stderr, "output: mising axis parameter\n");    
                exit(1);
            } 
            output_x[0][output_num]=x1;            
            output_x[1][output_num]=x2;            
            output_y[0][output_num]=y1;            
            output_y[1][output_num]=y2;            
            output_z[0][output_num]=z1;            
            output_z[1][output_num]=z2;  
            output_axis[output_num]=axis;
            strcpy(output_file[output_num], filename);
            output_num++;
            continue;
        }    
   
        if( !strcmp(type, "default_out") ) 
        {
            char filename[20];

            printf("\tdefault_out\n");       
            flag_default_out = 1;    /* set the flag to true */
            if( fscanf(InF, "%s", filename)!=1 ) 
            {
                fprintf(stderr, "default_out: mising output file name\n");
                exit(1);
            }
            strcpy(default_out_file, filename);
            continue;
        }
       
  
        fprintf(stderr,"error: Unrecognized keyword %s\n", type);
        exit(1);
    } /* end of sscanf(***, type ) */
 
    if( freq==0 ) 
       fprintf(stderr, "WARNING: input file does not have a source!\n\n");


    /* detect conductative boundary and inner conductors */ 
    for(i=0; i<CondNum; i++) 
    {
        x1=conductor[i].x1;
        y1=conductor[i].y1;
        z1=conductor[i].z1;
        x2=conductor[i].x2;
        y2=conductor[i].y2;
        z2=conductor[i].z2;
   
        /* conductative boundary maybe exist */
        if( ((x1==x2) && ((x1==XBoundDim[0]) || (x1==XBoundDim[1])))  ||
            ((y1==y2) && ((y1==YBoundDim[0]) || (y1==YBoundDim[1])))  ||
            ((z1==z2) && ((z1==ZBoundDim[0]) || (z1==ZBoundDim[1]))) )
        {
            DetectCondBound(conductor[i]);
        }
      
        /* internal conductor maybe exist */
        else if( ((x1==x2) && (x1>XBoundDim[0]) && (x1<XBoundDim[1])) ||
                 ((y1==y2) && (y1>YBoundDim[0]) && (y1<YBoundDim[1]))  ||
                 ((z1==z2) && (z1>ZBoundDim[0]) && (z1<ZBoundDim[1])) ) 
        {
            DetectInnerCond(conductor[i]);
        }
        else 
        {
            /* it's a external conductor */	
            Xextcond[0][ExtcondNum]= conductor[i].x1; 
            Xextcond[1][ExtcondNum]= conductor[i].x2;
            Yextcond[0][ExtcondNum]= conductor[i].y1; 
            Yextcond[1][ExtcondNum]= conductor[i].y2;
            Zextcond[0][ExtcondNum]= conductor[i].z1; 
            Zextcond[1][ExtcondNum]= conductor[i].z2;
            DelXextcond[ExtcondNum]= ( int ) conductor[i].DelX;
            DelYextcond[ExtcondNum]= ( int ) conductor[i].DelY;
            DelZextcond[ExtcondNum]= ( int ) conductor[i].DelZ;
            ExtcondNum++;
        }
    }       	
         	
    rewind(InF);
    OperateFreq = freq;
    CellDimension = CellDim;

}





/*************************************************************************
Prototype:    void    ComputeEdgeMultiplicity() 
Description:    To compute the multiplicity of edges. 
Input value:     none 
Return value:     none 
Global value used:     TotEdgeNum, EdgeStat, MetalEdgeStat, GlobalEdgeEnds,
       TotTrngleNum, PlusTrngleIndex, TrngleEdge, PlusTrngleDetect,      
       MinusTrngleIndex, MinusTrngleDetect, TotBoundEdgeNum, EdgeFound,   
       MultFound, VSourceNum, VSourceEdge, OperateFreq, E_Theta, E_Phi, 
       K_Theta, K_Phi, PlaneWaveE_Mag, ISourceEdgeStat, ISourceEdgeMag,   
       JunctionNum, Junction, flag_default_out, 
       output_num, output_file, OutFilePntr. 
Global value modified:    TotBoundEdgeNum 
Subroutines called:     none 
**************************************************************************/
void ComputeEdgeMultiplicity()
{

    int k,l, counter;

    for(ii=0;ii<TotEdgeNum;ii++)
    {
        /* metal boundary edges included by hybrid boundary */
        if( (EdgeStat[ii]==-1) && (MetalEdgeStat[ii]==1) )   
            GlobalEdgeEnds[ii][4]=1;

        /* metal edges which are not belong to the hybrid boundary */
        if( (EdgeStat[ii]==-1) && (MetalEdgeStat[ii]==2) ) 
            GlobalEdgeEnds[ii][4]=2;
 
    }



    /* detect edge multiplicity, also find out plus-minus triangle pairs */ 
    for(ii=0;ii<TotTrngleNum;ii++)
        for(i=0;i<3;i++)
        {
            /* counter how many triangles linked to 
               edge TrngleEdge[ii][i] */

            GlobalEdgeEnds[abs(TrngleEdge[ii][i])-1][2] +=1;

            /* find a plus-triangle */
            if( (TrngleEdge[ii][i])>0 ) 
            {
                /* jj is the number of plus trianges already found */
                jj = PlusTrngleIndex[abs(TrngleEdge[ii][i])-1];
                PlusTrngleDetect[abs(TrngleEdge[ii][i])-1][jj]=ii+1;
                PlusTrngleIndex[abs(TrngleEdge[ii][i])-1]++;
            }  /* if( TetFace */
 
            /* it is minus-triangle */
            else 
            {
                jj = MinusTrngleIndex[abs(TrngleEdge[ii][i])-1];
                MinusTrngleDetect[abs(TrngleEdge[ii][i])-1][jj]=ii+1;
                MinusTrngleIndex[abs(TrngleEdge[ii][i])-1]++;
            }  /* else */
    } /* end of for(..for */
    

    for(k=1; k<TotEdgeNum+1; k++) 
        if( GlobalEdgeEnds[k-1][4]>1 )  /* it's an external metal edge */
            for(l=1; l<TotEdgeNum+1; l++)       
                /* hybrid contour edge l has the same coordinates as 
                   edge k that means find a junction */
                if( GlobalEdgeEnds[l-1][4]==1 &&  (fabs((double)NodeCord[GlobalEdgeEnds[l-1][0]-1][0]-NodeCord[GlobalEdgeEnds[k-1][0]-1][0])<1e-5) &&
             (fabs((double)NodeCord[GlobalEdgeEnds[l-1][0]-1][1]-NodeCord[GlobalEdgeEnds[k-1][0]-1][1])<1e-5)&&
             (fabs((double)NodeCord[GlobalEdgeEnds[l-1][0]-1][2]-NodeCord[GlobalEdgeEnds[k-1][0]-1][2])<1e-5) &&
             (fabs((double)NodeCord[GlobalEdgeEnds[l-1][1]-1][0]-NodeCord[GlobalEdgeEnds[k-1][1]-1][0])<1e-5) &&
             (fabs((double)NodeCord[GlobalEdgeEnds[l-1][1]-1][1]-NodeCord[GlobalEdgeEnds[k-1][1]-1][1])<1e-5) &&
             (fabs((double)NodeCord[GlobalEdgeEnds[l-1][1]-1][2]-NodeCord[GlobalEdgeEnds[k-1][1]-1][2])<1e-5))          
                {
             
                    /* there are two triangels attached to this edge */
                    GlobalEdgeEnds[k-1][2]=2;
                
                    /* it already has a T-, thus need a T+ */
                    if( MinusTrngleIndex[k-1] ) 
                    {
                        PlusTrngleDetect[k-1][0]=PlusTrngleDetect[l-1][0];
                        PlusTrngleIndex[k-1]++;
                     
                        /* get a T++ junction */
                        Junction[JunctionNum][0]=l;
                        Junction[JunctionNum][1]=k;
                        Junction[JunctionNum][2]= 
                            PlusTrngleDetect[l-1][0]-1;
                        JunctionNum++;
                    }
                
                    /*it already  has a T+, thus need a T- */
                    else
                    {
                        MinusTrngleDetect[k-1][0]=
                            MinusTrngleDetect[l-1][0];
                        MinusTrngleIndex[k-1]++;
                          
                        /* get a T-- junction */
                        Junction[JunctionNum][0]=l;
                        Junction[JunctionNum][1]=k;
                        Junction[JunctionNum][2]=
                            MinusTrngleDetect[l-1][0]-1;
                        JunctionNum++;
                    }
                }   /* end of above 4 fold loop */
                    
               

    /* update total TotTrngleNum and TotEdgeNum */

    /* only edges having no less than two triangles linked with 
       them will be   included in C matrix. */
    for(ii=0;ii<TotEdgeNum;ii++) 
        if( GlobalEdgeEnds[ii][2]!=0 ) GlobalEdgeEnds[ii][2] -=1;

    for(ii=0;ii<TotEdgeNum;ii++) 
        TotBoundEdgeNum +=GlobalEdgeEnds[ii][2];


    /* those contour edges of external metal conductors */
    for(ii=0;ii<TotEdgeNum;ii++) 
        if( (PlusTrngleIndex[ii]==0) || (MinusTrngleIndex[ii]==0) ) 
            MetalEdgeStat[ii]=3;


    /* detect all internal and non-metal boundary edges */
    counter=0;
    for(ii=0;ii<TotEdgeNum;ii++)         
        if( (GlobalEdgeEnds[ii][4]==0) && (GlobalEdgeEnds[ii][2]>0) )  
        {
            /* those dielectric boundary edges should be included 
               once in  C matrix GlobalEdgeEnds[ii][4]==0  means 
               it is a dielectric edge. GlobalEdgeEnds[ii][2]>0 means 
               it is a boundary edge. GlobalEdgeEnds[ii][2]==0 for
               those inner edges  */

            GlobalEdgeEnds[ii][3]=counter;
            counter++;
        }


    /*detect all metal edges included by the hybrid boundary*/
    for(ii=0;ii<TotEdgeNum;ii++)
        if( (GlobalEdgeEnds[ii][4]==1) && (GlobalEdgeEnds[ii][2]>0) )
            GlobalEdgeEnds[ii][3]=counter++;

    /*detect those metal edges not included by the hybrid boundary*/
    for(ii=0;ii<TotEdgeNum;ii++)
        if( (GlobalEdgeEnds[ii][4]==2)&&(GlobalEdgeEnds[ii][2]>0) )
            GlobalEdgeEnds[ii][3]=counter++;   

}   /* end of ComputeEdgeMultiplicity() */





/*********************************************************************
Prototype:    void    DetectVSourceEdges() 
Description:  To find edges coinciding with the source. 
Input value:     none 
Return value:     none 
Global value used:   XextVsrc,YextVsrc, ZextVsrc TotEdgeNum, ExtVsrcNum,
    NodeCord, GlobalEdgeEnds, VSourceEdge,VSourceNum 
Global value modified:     none 
Subroutines called:     none 
**********************************************************************/
void DetectVSourceEdges()
{
    double NodeX[2],NodeY[2],NodeZ[2];
    int DelXextVsrc,DelYextVsrc, DelZextVsrc, flag=0;;

    for(IEdge=0;IEdge<ExtVsrcNum;IEdge++)
    {
        /* find out  DelXextVsrc,DelYextVsrc, DelZextVsrc first */
        flag=0;
        for(i=0; i<ExtcondNum; i++)
            if( XextVsrc[0][IEdge]>= Xextcond[0][i] && 
                XextVsrc[1][IEdge]<= Xextcond[1][i] &&
                YextVsrc[0][IEdge]>= Yextcond[0][i] && 
                YextVsrc[1][IEdge]<= Yextcond[1][i] &&
                ZextVsrc[0][IEdge]>= Zextcond[0][i] && 
                ZextVsrc[1][IEdge]<= Zextcond[1][i] )
            {
                DelXextVsrc=DelXextcond[i];
                DelYextVsrc=DelYextcond[i];
                DelZextVsrc=DelZextcond[i];
                flag=1;
                break;
            }
      
    
        if( flag==0 ) 
        {   
            fprintf(stderr, "Can't find vsource edge\n");
            fprintf(stderr, "Please check conductor and vsource\
keyowrds\n"); 
                      
            exit(1);
    
        }   
 
        for(i=XextVsrc[0][IEdge]; i<=XextVsrc[1][IEdge]; i+=DelXextVsrc)
            for(j=YextVsrc[0][IEdge]; j<=YextVsrc[1][IEdge]; 
                     j+= DelYextVsrc)
                for(k=ZextVsrc[0][IEdge]; k<=ZextVsrc[1][IEdge]; 
                       k+= DelZextVsrc)
                {

                    NodeX[0] = i; 
                    if( XextVsrc[1][IEdge]>XextVsrc[0][IEdge] )  
                        NodeX[1] = i + DelXextVsrc;
                    else  NodeX[1] = i;
                    NodeY[0] = j; 

                    if( YextVsrc[1][IEdge]>YextVsrc[0][IEdge] )  
                        NodeY[1] = j + DelYextVsrc;
                    else  NodeY[1] = j;

                    NodeZ[0] = k; 
                    if( ZextVsrc[1][IEdge]>ZextVsrc[0][IEdge] )  
                        NodeZ[1] = k + DelZextVsrc;
                    else  NodeZ[1] = k;


                    /* check whether the edge is out of 
                       the vsource edge boundary */
                    if( (NodeX[0]>=XextVsrc[0][IEdge]) && 
                        (NodeX[1]<=XextVsrc[1][IEdge]) &&
                        (NodeY[0]>=YextVsrc[0][IEdge]) &&  
                        (NodeY[1]<=YextVsrc[1][IEdge]) &&
                        (NodeZ[0]>=ZextVsrc[0][IEdge]) &&  
                        (NodeZ[1]<=ZextVsrc[1][IEdge])  )
    
                    {
 
                        NodeX[0] = CellDimension*NodeX[0];
                        NodeX[1] = CellDimension*NodeX[1];
                        NodeY[0] = CellDimension*NodeY[0];
                        NodeY[1] = CellDimension*NodeY[1];
                        NodeZ[0] = CellDimension*NodeZ[0];
                        NodeZ[1] = CellDimension*NodeZ[1];

                        for(JEdge=0;JEdge<TotEdgeNum;JEdge++)
                           if(
( (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][0]-1][0]-NodeX[0])<1E-5) &&
  (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][0]-1][1]-NodeY[0])<1E-5) &&
  (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][0]-1][2]-NodeZ[0])<1E-5) &&
  (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][1]-1][0]-NodeX[1])<1E-5) &&
  (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][1]-1][1]-NodeY[1])<1E-5) &&
  (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][1]-1][2]-NodeZ[1])<1E-5) 
) ||
  ( (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][1]-1][0]-NodeX[0])<1E-5) &&
  (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][1]-1][1]-NodeY[0])<1E-5) &&
  (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][1]-1][2]-NodeZ[0])<1E-5) &&
  (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][0]-1][0]-NodeX[1])<1E-5) &&
  (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][0]-1][1]-NodeY[1])<1E-5) &&
  (fabs((double)NodeCord[GlobalEdgeEnds[JEdge][0]-1][2]-NodeZ[1])<1E-5)
))
    
                           {
 
                               VSourceEdge[VSourceNum]=JEdge+1;  
                               VSourceMag[VSourceNum]=ExtVsrcMag[IEdge];
                               VSourceNum++;
      
                           }
                   }      
               }         
        }    


        /* check whether those source edges are defined on those
           extented path edge */


        for(IEdge=0;IEdge<VSourceNum;IEdge++)
        {
            /* indicate whether the source edge is in BoundEdgeState */
            flag=0;         
 
    	    for(JEdge=0;JEdge<TotBoundEdgeNum;JEdge++)                    
                if( BoundEdgeStat[JEdge]==VSourceEdge[IEdge] ) 
                {
                    /* vsource can not be defined on the hybridboudary */
                    if( JEdge<HybrdBoundEdgeNum )
                    {
                        fprintf(stderr, "vsource can not be defined \
on the hybrid boudary\n!");
                        exit(1);

                    }  /* if( JEdge */

                flag=1;
                break;
 
        }  /* if(BoundEdgeStat[JEdge] */

         
        /*Can't find the Vsource edge */
        if( flag==0 ) 
        {
            fprintf(stderr, "Check your Vsource edge\n");
            exit(1);
        } /* if(flag) */

    }  /* for(IEdge=0;  */

    if( VSourceNum==0 )
    {
        fprintf(stderr,"No source edge found!\n");      
        exit(1);
    }     


}   /* end of DetectVSourceEdges() */
  


/*  return the maxium of two integer numbers */
int Max(int x1, int x2)
{ 
   return (x1>x2?x1:x2);
}



/* return the minum of two integer numbers */
int Min(int x1, int x2)
{ 
   return (x1<x2?x1:x2);
}




/**************************************************************************** 
Prottotype:  int Intersection(RectAngle R1, RectAngle R2, RectAngle *R, 
                 int *SubRectNum, RectAngle *SubRect)
Description:
    detect intersection between two rectangles R1 and R2.
    If return value equal to 1, then an  intersection is found and stored 
    in R. Otherwise return 0. SubRectNum returns the number of rectangles R2
    is devided by R1. sub-area stores the rectangle information.
Input value:  
    RectAngle R1, R2 ---- the two rectanges to be detected whether
                          there are intersections
    RectAngle R  --- to store the intersections
    int *SubRectNum --- to store the other rectangles formed after the  
                        intersection area is taken from R2.
    RectAngle SubRect ---  store the coordinates of rectangles in R2
                           divided by R1 except the intersection area.
Output value:
    If any intersection is found, return 1. Otherwise return 0. 
Global value used:  none
Global value modified: none
****************************************************************************/
int Intersection(RectAngle R1, RectAngle R2, RectAngle *R, 
                 int *SubRectNum, RectAngle *SubRect)
{
    int flag=0;
 
 
    *SubRectNum=0;

    if( (R2.end.x<= R1.start.x) ||(R2.start.x>=R1.end.x) ||
        (R2.end.y<= R1.start.y) ||(R2.start.y>=R1.end.y) )
    {
    
    	/* the sub-rectangle is R2 itself */
        SubRect[*SubRectNum].start.x= R2.start.x;
        SubRect[*SubRectNum].start.y=R2.start.y;
   	SubRect[*SubRectNum].end.x=R2.end.x;
   	SubRect[*SubRectNum].end.y=R2.end.y;
        (*SubRectNum)++;
    }
          
    else
    {
   	R->start.x=Max(R2.start.x, R1.start.x );
   	R->start.y=Max(R2.start.y, R1.start.y);
   	R->end.x=Min(R2.end.x, R1.end.x);
   	R->end.y=Min(R2.end.y, R1.end.y);
   	
   	flag=1;
   	
   	if( R->start.y>R2.start.y)
        {
   	    SubRect[*SubRectNum].start.x= R2.start.x;
   	    SubRect[*SubRectNum].start.y=R2.start.y;
   	    SubRect[*SubRectNum].end.x=R2.end.x;
   	    SubRect[*SubRectNum].end.y=R->start.y;
   	    (*SubRectNum)++;
   	}
   	  
   	if( R2.end.y>R->end.y )
        {
    	    SubRect[*SubRectNum].start.x= R2.start.x;
   	    SubRect[*SubRectNum].start.y=R2.start.y;
   	    SubRect[*SubRectNum].end.x=R2.end.x;
   	    SubRect[*SubRectNum].end.y=R->start.y;
   	    (*SubRectNum)++;
   	}
   	  
   	if( R2.end.y>R->end.y )
        {
   	    SubRect[*SubRectNum].start.x=R2.start.x ;
   	    SubRect[*SubRectNum].start.y=R->end.y;
   	    SubRect[*SubRectNum].end.x=R2.end.x;
   	    SubRect[*SubRectNum].end.y=R2.end.y;
   	    (*SubRectNum)++;
   	}
   	 
   	
     	if( R->start.x>R2.start.x )
        {
   	    SubRect[*SubRectNum].start.x= R2.start.x;
   	    SubRect[*SubRectNum].start.y=R->start.y;
   	    SubRect[*SubRectNum].end.x=R->start.x;
   	    SubRect[*SubRectNum].end.y=R->end.y;
  	    (*SubRectNum)++;
   	}
   	
  	
    	if( R2.end.x>R->end.x )
        {
   	    SubRect[*SubRectNum].start.x=R->end.x ;
   	    SubRect[*SubRectNum].start.y=R->start.y;
   	    SubRect[*SubRectNum].end.x=R2.end.x;
   	    SubRect[*SubRectNum].end.y=R->end.y;
   	    (*SubRectNum)++;
   	}

    }  /* else */
    
    return flag;

}  /* end of Intersection()  */




/************************************************************************
Prototype:    void    PrintOutput() 
Description:  To print out the mesh data, source information and output   
              specifications to a file. This file will be used by emap5. 
Input value:     none 
Return value:     none 
Global value used:   TotEdgeNum, TotTetElement,   
                     TotNodeNum, TotTrngleNum, TotBoundEdgeNum, 
                     TotInnerEdgeNum, DielBoundEdgeNum, 
                     TotISourceEdgeNum,TrngleNode, TrngleEdge, NodeCord, 
                     TetNode, MediaPerm, TetEdge, GlobalEdgeEnds, 
                     PlusTrngleIndex, PlusTrngleDetect, MinusTrngleIndex, 
                     MinusTrngleDetect, InnerEdgeStat, BoundaryEdgeStat, 
                     VsourceMag, VsourceNum, OperateFreq, 
Global value modified:     none 
Subroutines called:     none 
*************************************************************************/
void PrintOutput()
{

    fprintf(OutFilePntr, "# Global Information Table\n");
    fprintf(OutFilePntr,"%d\t",TotEdgeNum);
    fprintf(OutFilePntr,"%d\t",TotTetElement);
    fprintf(OutFilePntr,"%d\t",TotNodeNum);
    fprintf(OutFilePntr,"%d\n",TotTrngleNum);
 
    fprintf(OutFilePntr,"%d\t",TotBoundEdgeNum);
    fprintf(OutFilePntr,"%d\t",TotInnerEdgeNum);
    fprintf(OutFilePntr,"%d\t",DielBoundEdgeNum);
    fprintf(OutFilePntr,"%d\n",HybrdBoundEdgeNum);

    fprintf(OutFilePntr,"\n\n"); 
 
    fprintf(OutFilePntr, "# Global Node Table\n");
    for(ii=0;ii<TotNodeNum;ii++) 
    {
        for(jj=0;jj<=2;jj++) 
        fprintf(OutFilePntr,"%lf\t",NodeCord[ii][jj]);
        fprintf(OutFilePntr,"\n");
    }

    fprintf(OutFilePntr,"\n\n# Global Edge Table\n");
    for(ii=0;ii<TotEdgeNum;ii++)
    {
        for(jj=0;jj<6;jj++) 
            fprintf(OutFilePntr,"%d\t",GlobalEdgeEnds[ii][jj]);
        fprintf(OutFilePntr,"\n");
    }
    
    fprintf(OutFilePntr,"\n\n# Triangle Table\n");
    for(jj=0;jj<TotTrngleNum;jj++)
    {
        for(j=0;j<=2;j++)
            fprintf(OutFilePntr,"%d\t",TrngleNode[jj][j]);
        for(j=0;j<=2;j++)
            fprintf(OutFilePntr,"%d\t",TrngleEdge[jj][j]);
        fprintf(OutFilePntr,"\n");
    }
        
    fprintf(OutFilePntr,"\n# Tetrahedron Table\n");
    for(ii=0;ii<TotTetElement;ii++) 
    { 
        for(jj=0;jj<=3;jj++)
            fprintf(OutFilePntr,"%d\t",TetNode[ii][jj]);
        fprintf(OutFilePntr,"%lf\t%lf\t",MediaPerm[ii][0],MediaPerm[ii][1]);
        fprintf(OutFilePntr,"\n"); 
    }

    fprintf(OutFilePntr,"\n"); 

    for(ii=0;ii<TotTetElement;ii++) 
    {
        for(jj=0;jj<=5;jj++)
            fprintf(OutFilePntr,"%d\t",TetEdge[ii][jj]);  
        fprintf(OutFilePntr,"\n");
    }

    fprintf(OutFilePntr,"\n\n");

    fprintf(OutFilePntr,"# Plus/Minus Triangle Table\n");
 
    for(ii=0;ii<TotEdgeNum;ii++)
    {
        fprintf(OutFilePntr,"%d\t",PlusTrngleIndex[ii]);

        if (PlusTrngleIndex[ii]==0)  fprintf(OutFilePntr,"\t");
        else 
        {   for(i=0;i<PlusTrngleIndex[ii];i++)
                fprintf(OutFilePntr,"%d\t  ",PlusTrngleDetect[ii][i]);
        }
     
        fprintf(OutFilePntr,"%d\t\t",MinusTrngleIndex[ii]);

        for(i=0;i<MinusTrngleIndex[ii];i++)
            fprintf(OutFilePntr,"%d\t",MinusTrngleDetect[ii][i]);

        fprintf(OutFilePntr,"\n");
    }   /* Plus/minus table  */
 
    fprintf(OutFilePntr,"\n\n");
 
    fprintf(OutFilePntr,"# Inner Edge Table\n");
    for(ii=0;ii<TotInnerEdgeNum;ii++)
    {
        if(ii%8==0)fprintf(OutFilePntr,"\n");
        fprintf(OutFilePntr,"%d\t",InnerEdgeStat[ii]);
    }  /* InnerEdge table */


    fprintf(OutFilePntr, "\n\n");

    fprintf(OutFilePntr, "# Boundary Edge Table\n");
    for(ii=0;ii<TotBoundEdgeNum;ii++)
    {
        if(ii%8==0 && ii!=0) fprintf(OutFilePntr,"\n");
        fprintf(OutFilePntr,"%d\t",BoundEdgeStat[ii]);
    }

    fprintf(OutFilePntr, "\n\n");

    fprintf(OutFilePntr, "# Source Information\n");
    if( SourceType=='V' )
    {
        fprintf(OutFilePntr,"%s\t%lf\t%d\n","V", OperateFreq, VSourceNum);
                           
        for(ii=0;ii<VSourceNum;ii++)
            fprintf(OutFilePntr,"%d\t%lf\n",VSourceEdge[ii], VSourceMag[ii]);
 
    }  /* if( SourceType=='V' */   

    /* A plane wave incides upon the structure */
    else  if( SourceType=='P' )    
              fprintf(OutFilePntr,"%s\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
                      "P", OperateFreq,E_Theta,E_Phi,K_Theta, 
                       K_Phi,PlaneWaveE_Mag);
             
    else  
    {
        /* isource */

        fprintf(OutFilePntr,"%s %lf %d\n", "I", OperateFreq,
                TotISourceEdgeNum );

        for(ii=0; ii<TotISourceEdgeNum; ii++)
             fprintf(OutFilePntr,"%d  %lf\n",ISourceEdgeStat[ii], 
                     ISourceEdgeMag[ii]);
   
    }  /* else */

    fprintf(OutFilePntr, "\n\n" );


    fprintf(OutFilePntr, "# Junction Table\n" );
    fprintf(OutFilePntr,"%d\n", JunctionNum);
    for(ii=0; ii<JunctionNum; ii++)
        fprintf(OutFilePntr,"%d  %d  %d\n",   Junction[ii][0], 
                            Junction[ii][1],Junction[ii][2]);


    fprintf(OutFilePntr, "\n\n# Resistor Table\n" );
    fprintf(OutFilePntr, "resistor");
    {
        int ResistorEdgeStat[100];
        double ResistorEdgeValue[100];
        int ResistorEdgeNum=0;

        for(ii=0; ii<ResisterNum; ii++) 
        {
            int i;

            for(i=0; i<TotInnerEdgeNum; i++)
            {
                double start[3], end[3];
                int counter;
 
                for(counter=0; counter<3; counter++)
                {            start[counter]=NodeCord[GlobalEdgeEnds[InnerEdgeStat[i]-1][0]-1][counter];
end[counter]=NodeCord[GlobalEdgeEnds[InnerEdgeStat[i]-1][1]-1][counter];
                }

                if( (start[0]+1.0e-6)>(ResisterLocX[ii][0]*CellDimension) &&
                    (start[0]-1e-6)<(ResisterLocX[ii][1]*CellDimension) &&
	            (start[1]+1e-6)>(ResisterLocY[ii][0]*CellDimension) &&
	            (start[1]-1e-6)<(ResisterLocY[ii][1]*CellDimension) &&
	            (start[2]+1e-6)>(ResisterLocZ[ii][0]*CellDimension) &&
	            (start[2]-1e-6)<(ResisterLocZ[ii][1]*CellDimension) &&
	            (end[0]+1e-6)>(ResisterLocX[ii][0]*CellDimension) &&
                    (end[0]-1e-6)<(ResisterLocX[ii][1]*CellDimension) &&
	            (end[1]+1e-6)>(ResisterLocY[ii][0]*CellDimension) &&
	            (end[1]-1e-6)<(ResisterLocY[ii][1]*CellDimension) &&
	            (end[2]+1e-6)>(ResisterLocZ[ii][0]*CellDimension) &&
	            (end[2]-1e-6)<(ResisterLocZ[ii][1]*CellDimension) )
                {    

                    ResistorEdgeStat[ResistorEdgeNum]=InnerEdgeStat[i];
                    ResistorEdgeValue[ResistorEdgeNum++]=ResisterValue[ii];
           

                }  /*  if()  */
            }      /*  for(i=0 */
        }          /*  for ii=0 */

        fprintf(OutFilePntr, "\t%d\n", ResistorEdgeNum);
        for(ii=0; ii<ResistorEdgeNum; ii++)
            fprintf(OutFilePntr, "%d %f\n", ResistorEdgeStat[ii],    
                    ResistorEdgeValue[ii]);
    }
     
    fprintf(OutFilePntr, "\n\n");
      
    fprintf(OutFilePntr, "# Output Table\n");
    fprintf(OutFilePntr, "default_out\t%d\t", flag_default_out);
 
    if (flag_default_out==1) fprintf(OutFilePntr,"%s", default_out_file );

    fprintf(OutFilePntr, "\n\n");
 
    /* output */
    fprintf(OutFilePntr, "%d\n", output_num);

    for(ii=0; ii<output_num; ii++) 
    {

        double MinPoint[3], MaxPoint[3];
     
        MinPoint[0] = CellDimension * output_x[0][ii] - 1e-5;
        MinPoint[1] = CellDimension * output_y[0][ii] - 1e-5;
        MinPoint[2] = CellDimension * output_z[0][ii] - 1e-5;
        MaxPoint[0] = CellDimension * output_x[1][ii] + 1e-5;
        MaxPoint[1] = CellDimension * output_y[1][ii] + 1e-5;
        MaxPoint[2] = CellDimension * output_z[1][ii] + 1e-5;

        fprintf(OutFilePntr, "%s\n", output_file[ii]);
        fprintf(OutFilePntr, "%lf %lf %lf %lf %lf %lf %c\n\n", 
                CellDimension * output_x[0][ii],
                CellDimension * output_y[0][ii],
                CellDimension * output_z[0][ii],
                CellDimension * output_x[1][ii],
                CellDimension * output_y[1][ii],
                CellDimension * output_z[1][ii],
                output_axis[ii] ); 

    }          
 
    fprintf(OutFilePntr,"#  End of the Input File\n");
    fclose(OutFilePntr);

    printf("\nInput File Report:\n\n");
    printf("\tFEM Volume:\n");
    printf("\t    segments along the x axis:  %d\n", XdiM);
    printf("\t    segments along the y axis:  %d\n", YdiM);
    printf("\t    segments along the z axis:  %d\n", ZdiM);
    printf("\t    total inner edges:          %d\n", TotInnerEdgeNum);
    printf("\t    total tetrahetron elements: %d\n", TotTetElement);
    printf("\n\tMoM Region:\n");
    printf("\t    total triangle elements:        %d\n", TotTrngleNum);
    printf("\t    total boundary edges:           %d\n", TotBoundEdgeNum);
    printf("\t    total metallic boundary edges:  %d\n", 
            TotBoundEdgeNum-HybrdBoundEdgeNum); 
    printf("\t    total junction edges:           %d\n\n", JunctionNum );

    if (SourceType=='V')
        printf("\tnum of voltage source edges:    %d\n\n", VSourceNum);
    if (SourceType=='I')
        printf("\tnum of current source edges:    %d\n\n", TotISourceEdgeNum);

    printf("Memory Required:   %7.2f M Bytes\n",
                2*sizeof(double)*( pow((double)TotBoundEdgeNum,2.0) + 
                                   TotBoundEdgeNum*HybrdBoundEdgeNum+ 
                                   +11*TotInnerEdgeNum )/1.0e6  );    
    

}  /* end of PrintOutput() */





/***************************************************************************
Prototype:    void    DiscretizeFEMVolume() 
Description:  To disretize the FEM volume into tetrahedra. 
Input value:     none 
Return value:     none 
Global value used:   CellUnitXNum, CellUnitYNum, CellUnitZNum, 
    MediaPerm, FaceNode, FaceEdge, TrngleNode, TrngleEdge, 
Global value modified:     none 
Subroutines called:     TetraEdgeDiv1(), GlobalEdgeEndsDiv1(), 
    TetraEdgeDiv2(), GlobalEdgeEndsDiv2(),  FindTetraHedronFaceNodes(),  
    ComputeTrngleCentroid(; ModifyBoundFaces();
***************************************************************************/
void DiscretizeFEMVolume()
{
    int IEdge, i, j, k, ii, con, XmaxDiel,YmaxDiel,ZmaxDiel,
        XminDiel,YminDiel,ZminDiel;

    /* begin to discretize the dielectric boxes */
    for(IEdge=0;IEdge<DielmatNum;IEdge++) 
    {
        XmaxDiel=YmaxDiel= ZmaxDiel=0;
        XminDiel=YminDiel= ZminDiel=0;
   
        /* calculate XminDiel */
        for(i=0; i<CellUnitXNum; i++)
        if( fabs((double)XDielmat[0][IEdge]-CellUnitX[i].start)<1e-5 )break;
      
        if( i==CellUnitXNum )
        {
            fprintf(stderr,"Please check Dielectric box dimension\n");
            exit(1);
        }
   
        for(j=0; j<i; j++)
            XminDiel+= (int ) ( fabs((double)CellUnitX[j].end-CellUnitX[j].start) 
                      /CellUnitX[j].step+1e-3 );
     
 
        /* calculate XmaxDiel */
     
        for(i=0; i<CellUnitXNum; i++)
            if(fabs((double)XDielmat[1][IEdge]-CellUnitX[i].end)<1e-5) break;

        if( i==CellUnitXNum )
        {
            fprintf(stderr,"Please check Dielectric box dimension\n");
            exit(1);
        }
   
        for(j=0; j<=i; j++)
            XmaxDiel+= (int ) ( fabs((double)CellUnitX[j].end-CellUnitX[j].start)
                   /CellUnitX[j].step+1e-3 );
 
     
        /* calculate YminDiel */
        for(i=0; i<CellUnitYNum; i++)
            if( fabs((double)YDielmat[0][IEdge]-CellUnitY[i].start)<1e-5 ) break;

        if( i==CellUnitYNum )
        {
            fprintf(stderr,"Please check Dielectric box dimension\n");
            exit(1);
        }
   
        for(j=0; j<i; j++)
            YminDiel += (int ) (fabs((double)CellUnitY[j].end-CellUnitY[j].start)
                        /CellUnitY[j].step+1e-3 );
     
 
        /* calculate YmaxDiel */
        for(i=0; i<CellUnitYNum; i++)
            if( fabs((double)YDielmat[1][IEdge]-CellUnitY[i].end)<1e-5 ) break;

        if( i==CellUnitYNum )
        {
            fprintf(stderr,"Please check Dielectric box dimension\n");
            exit(1);
        }
   
        for(j=0; j<=i; j++)
            YmaxDiel += (int) ( fabs((double)CellUnitY[j].end-CellUnitY[j].start)
                        /CellUnitY[j].step+1e-3 );
     
     
        /* calculate ZminDiel */
        for(i=0; i<CellUnitZNum; i++)
            if( fabs((double)ZDielmat[0][IEdge]-CellUnitZ[i].start)<1e-5 ) break;

        if( i==CellUnitZNum ) 
        {
            fprintf(stderr,"Please check Dielectric box dimension\n");
            exit(1);
        }
   
        for(j=0; j<i; j++)
            ZminDiel+= (int) ( fabs((double)CellUnitZ[j].end-CellUnitZ[j].start)  
                      /CellUnitZ[j].step+1e-3 );
     
 
        /* calculate ZmaxDiel */
        for(i=0; i<CellUnitZNum; i++)
            if( fabs((double)ZDielmat[1][IEdge]-CellUnitZ[i].end)<1e-5 ) break;

        if( i==CellUnitZNum )
        {
            fprintf(stderr,"Please check Dielectric box dimension\n");
            exit(1);
        }
   
        for(j=0; j<=i; j++)
            ZmaxDiel += (int) ( fabs((double)CellUnitZ[j].end-CellUnitZ[j].start) 
                        /CellUnitZ[j].step+1e-3 );
     
   
        for(i=XminDiel;i<XmaxDiel;i++)
            for(j=YminDiel;j<YmaxDiel;j++)
                for(k=ZminDiel;k<ZmaxDiel;k++)
                {
                    HexNum=k*XmaxDiel*YmaxDiel+j*XmaxDiel+i;
                    for(ii=0;ii<=4;ii++) 
                    {
                        MediaPerm[HexNum*5+ii][0] = Epsilon[IEdge];
                        MediaPerm[HexNum*5+ii][1] = Sigma[IEdge];
                    } /* end for(ii */
                } /* end for(k */
    } /* end for(IEDGE  */
 

    for(k=0;k<ZdiM;k++)
        for(j=0;j<YdiM;j++)
            for(i=0;i<XdiM;i++)  
            {
                HexNum = k*YdiM*XdiM + j*XdiM + i;
                con = (int) ( pow(-1.0,(double)i)
                     *pow(-1.0,(double)j)*pow(-1.0,(double)k) );

                if( con==1 ) 
                {
                    TetraEdgeDiv1(); 
                    GlobalEdgeEndsDiv1();
                }  /* if(con==1 */
                else       
                {
                    TetraEdgeDiv2(); 
                    GlobalEdgeEndsDiv2();
                }  /* else */
            }  /* for(i=0 */

    TotTetFaceNum = 0;

    for(ii=0;ii<TotTetElement;ii++) 
        FindTetraHedronFaceNodes(ii);	

    for(ii=0;ii<TotTetFaceNum;ii++)   
        ModifyBoundFaces(ii);

    TrngleCenTroid = double_Matrix(TotTrngleNum,3);
    for(II=0;II<TotTetFaceNum;II++)
        ComputeTrngleCentroid(II,NodeCord,TrngleNode,TrngleCenTroid);


    kk=0;

    /***********Right wall (x=0)************/
    for(jj=0;jj<TotTetFaceNum;jj++)
        if( fabs((double)TrngleCenTroid[jj][0]-(double)(CellDimension*XBoundDim[0]))
            <1.0e-6 )
        {
            for(j=0;j<=2;j++) 
            {
                FaceNode[kk][j] = TrngleNode[jj][j];
                FaceEdge[kk][j] = TrngleEdge[jj][j]; 
            }
            kk++;
        }

    /***********Left wall (x=XdiM-1)********/
    for(jj=0;jj<TotTetFaceNum;jj++)    
       if( fabs((double)TrngleCenTroid[jj][0]-(double)(CellDimension*XBoundDim[1])) 
           <1.0e-6 )
       {
           for(j=0;j<=2;j++) 
           {
               FaceNode[kk][j] = TrngleNode[jj][j];
               FaceEdge[kk][j] = TrngleEdge[jj][j];
           }

           kk++;
       }   /* if() */
    

    /***********Bottom pec wall (y=0)*******/
    for(jj=0;jj<TotTetFaceNum;jj++)
        if( fabs((double)TrngleCenTroid[jj][1]-(double)(CellDimension*YBoundDim[0]))    
            <1.0e-6 )
        {
            for(j=0;j<=2;j++)
            {
                FaceNode[kk][j] = TrngleNode[jj][j];
                FaceEdge[kk][j] = TrngleEdge[jj][j];
            }
            kk++;
        }

    /**********Top pec wall (y=YdiM-1)******/
    for(jj=0;jj<TotTetFaceNum;jj++)
        if( fabs((double)TrngleCenTroid[jj][1]-(double)(CellDimension*YBoundDim[1]))
            <1.0e-6 )
        {
            for(j=0;j<=2;j++) 
            {
                FaceNode[kk][j] = TrngleNode[jj][j];
                FaceEdge[kk][j] = TrngleEdge[jj][j];
            }
            kk++;
        }

    /***********Front wall (z=0)************/
    for(jj=0;jj<TotTetFaceNum;jj++)
        if( fabs((double)TrngleCenTroid[jj][2]-(double)(CellDimension*ZBoundDim[0]))
            <1.0e-6 )
        {
            for(j=0;j<=2;j++) 
            {
               FaceNode[kk][j] = TrngleNode[jj][j];
               FaceEdge[kk][j] = TrngleEdge[jj][j];
            }

            kk++;
        }

    /***********Back wall (z=ZdiM-1)********/
    for(jj=0;jj<TotTetFaceNum;jj++)
        if( fabs((double)TrngleCenTroid[jj][2]-(double)(CellDimension*ZBoundDim[1]))  
            <1.0e-6 )
        {
            for(j=0;j<=2;j++) 
            {
                FaceNode[kk][j] = TrngleNode[jj][j];
                FaceEdge[kk][j] = TrngleEdge[jj][j];
            }
            kk++;
        }

    /*******************************/
    for(jj=0;jj<kk;jj++)
        for(j=0;j<=2;j++) {
            TrngleNode[jj][j] = FaceNode[jj][j];
            TrngleEdge[jj][j] = FaceEdge[jj][j];
        }  /* for(j=0 */



}   /* end of DiscretizeFEMVolume() */


/***************************************************************************
Prototype:    void    DiscretizeExterPECPatch() 
Description:  To disretize the external conductative surface into triangles. 
Input value:     none 
Return value:     none 
Global value used:  ExtcondNum, CubeNumNode, CubeTrnglleNum, DipoleWid, 
    DipolLen, CellDimension, DelXextcond,  DelYextcond, DelZextcond, 
    Dipolplane_XY, Dipolplane_YZ, Dipolplane_XZ, GlobalEdgeEnds 
Global value modified:     none 
Subroutines called:   none ***************************************************************************/
void DiscretizeExterPECPatch()
{
    int IEdge, i, j;
    int NodeOffset,EdgeOffset1,EdgeOffset2;

    CubeNumNode   = TotCubeNumNode;
    CubeNumEdge   = TotCubeNumEdge ;
    CubeTrngleNum = TotCubeTrngleNum;

    for(IEdge=0;IEdge<ExtcondNum;IEdge++)
    {
        for(i=0;i<=DipolLen[IEdge];i++)
        {
            KK = CubeNumNode + (DipolWid[IEdge]+1)*i;
            for(j=0;j<=DipolWid[IEdge];j++)
            {
                if( Dipolplane_XY[IEdge]==1 )
                { 
                    NodeCord[KK+j][0] = (double) 
                        (CellDimension*Xextcond[0][IEdge] 
                         + j*CellDimension*DelXextcond[IEdge]);
                    NodeCord[KK+j][1] = (double) 
                        (CellDimension*Yextcond[0][IEdge] 
                         + i*CellDimension*DelYextcond[IEdge]);
                    NodeCord[KK+j][2] = (double) 
                        (CellDimension*Zextcond[0][IEdge]);
                }  /* if( Dipole  */

                if( Dipolplane_YZ[IEdge]==1 )
                { 
                    NodeCord[KK+j][0] =  
                        (double)(CellDimension*Xextcond[0][IEdge]);
                    NodeCord[KK+j][1] = 
                        (double)(CellDimension*Yextcond[0][IEdge] 
                                     + j*CellDimension*DelYextcond[IEdge]);
                    NodeCord[KK+j][2] = 
                        (double)(CellDimension*Zextcond[0][IEdge] 
                                     + i*CellDimension*DelZextcond[IEdge]);
                }

                if( Dipolplane_ZX[IEdge]==1 )
                {
                    NodeCord[KK+j][0] = 
                        (double)(CellDimension*Xextcond[0][IEdge] 
                        + j*CellDimension*DelXextcond[IEdge]);
                    NodeCord[KK+j][1] = 
                        (double)(CellDimension*Yextcond[0][IEdge]);
                    NodeCord[KK+j][2] = 
                         (double)(CellDimension*Zextcond[0][IEdge] 
                                  + i*CellDimension*DelZextcond[IEdge]);
                } /* if(Dipolplane_ZX */
            } /* for(j=0;  */
        }  /* for(i=0;  */

        /******************************/
        for(i=0;i<DipolLen[IEdge];i++)
            for(j=0;j<DipolWid[IEdge];j++)
            {
   
                int flag;
    
                /* flag is used to define the normal vector for the
                   external conductor if there is no junction, the normal
                   vector doens't matter.if there do have junctions, the 
                   normal vector should be chosed carefully. Otherwise, 
                   there are problems assocaited with junction */

                if( Dipolplane_XY[IEdge]==1 )
                {
                    if( Zextcond[1][IEdge]>=ZBoundDim[1] ) flag=1;
                    else  flag=-1;
                }

                else if(Dipolplane_YZ[IEdge]==1)
                {
                    if( Xextcond[1][IEdge]>=XBoundDim[1] ) flag=1;
                    else  flag=-1;
                }

                else 
                {   /* parallel to XZ plane */
                    if(Yextcond[1][IEdge]>=YBoundDim[1]) flag=1;
                    else  flag=-1;
                }
  
                KK = CubeTrngleNum + 2*j + 2*DipolWid[IEdge]*i;
                NodeOffset = CubeNumNode + j + (DipolWid[IEdge]+1)*i;
                EdgeOffset1 = CubeNumEdge + j   + (3*DipolWid[IEdge]+1)*i;
                EdgeOffset2 = CubeNumEdge + 2*j   + (3*DipolWid[IEdge]+1)*i;
    
                /* revised here on Feb 12 */
                if( ((i%2==0) && (j%2==0)) || ((i%2==1) && (j%2==1)) )
                {
                    TrngleNode[KK+0][0] = NodeOffset + DipolWid[IEdge]+3;
                    TrngleNode[KK+0][1] = NodeOffset + DipolWid[IEdge]+2;
                    TrngleNode[KK+0][2] = NodeOffset + 1;
 
                    TrngleNode[KK+1][0] = NodeOffset + DipolWid[IEdge]+3;
                    TrngleNode[KK+1][1] = NodeOffset + 1;
                    TrngleNode[KK+1][2] = NodeOffset + 2;
 
                    TrngleEdge[KK+0][0] = -(EdgeOffset2 + 
                                             DipolWid[IEdge]+1)*flag;
                    TrngleEdge[KK+0][1] = (EdgeOffset2 + 
                                            DipolWid[IEdge]+2)*flag;
                    TrngleEdge[KK+0][2] = -(EdgeOffset1 + 
                                             3*DipolWid[IEdge]+2)*flag;
 
                    TrngleEdge[KK+1][0] = (EdgeOffset1 + 1)*flag;
                    TrngleEdge[KK+1][1] = (EdgeOffset2 +  
                                            DipolWid[IEdge]+3)*flag;
                    TrngleEdge[KK+1][2] = -(EdgeOffset2 + 
                                             DipolWid[IEdge]+2)*flag;

                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][0])-1][0] 
                         = NodeOffset + 1;
                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][0])-1][1] 
                         = NodeOffset + DipolWid[IEdge]+2;
                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][1])-1][0] 
                         = NodeOffset + 1;
                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][1])-1][1] 
                         = NodeOffset + DipolWid[IEdge]+3;
                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][2])-1][0] 
                         = NodeOffset + DipolWid[IEdge]+2;
                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][2])-1][1] 
                         = NodeOffset + DipolWid[IEdge]+3;
                    GlobalEdgeEnds[abs(TrngleEdge[KK+1][0])-1][0] 
                         = NodeOffset + 1;
                    GlobalEdgeEnds[abs(TrngleEdge[KK+1][0])-1][1] 
                         = NodeOffset + 2;
                    GlobalEdgeEnds[abs(TrngleEdge[KK+1][1])-1][0] 
                         = NodeOffset + 2;
                    GlobalEdgeEnds[abs(TrngleEdge[KK+1][1])-1][1] 
                         = NodeOffset + DipolWid[IEdge]+3;

                }   /* if() */
                else
                {
                    TrngleNode[KK+0][0] = NodeOffset + DipolWid[IEdge]+2;
                    TrngleNode[KK+0][1] = NodeOffset + 1;
                    TrngleNode[KK+0][2] = NodeOffset + 2;
 
                    TrngleNode[KK+1][0] = NodeOffset + DipolWid[IEdge]+2;
                    TrngleNode[KK+1][1] = NodeOffset + 2;
                    TrngleNode[KK+1][2] = NodeOffset + DipolWid[IEdge]+3;
 
                    TrngleEdge[KK+0][0] = (EdgeOffset1 + 1)*flag;
                    TrngleEdge[KK+0][1] = (EdgeOffset2 + 
                                            DipolWid[IEdge]+2)*flag;
                    TrngleEdge[KK+0][2] = -(EdgeOffset2 + 
                                             DipolWid[IEdge]+1)*flag;
 
                    TrngleEdge[KK+1][0] = (EdgeOffset2 + 
                                            DipolWid[IEdge]+3)*flag;
                    TrngleEdge[KK+1][1] =-(EdgeOffset1 + 
                                            3*DipolWid[IEdge]+2)*flag;
                    TrngleEdge[KK+1][2] =-(EdgeOffset2 + 
                                            DipolWid[IEdge]+2)*flag;

                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][0])-1][0] = 
                          NodeOffset + 1;
                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][0])-1][1] = 
                          NodeOffset + 2;
 
                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][1])-1][0] =
                          NodeOffset + 2;

                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][1])-1][1] = 
                          NodeOffset + DipolWid[IEdge]+2;
 
                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][2])-1][0] = 
                          NodeOffset + 1;
                    GlobalEdgeEnds[abs(TrngleEdge[KK+0][2])-1][1] =
                          NodeOffset + DipolWid[IEdge]+2;
 
                    GlobalEdgeEnds[abs(TrngleEdge[KK+1][0])-1][0] = 
                          NodeOffset + 2;

                    GlobalEdgeEnds[abs(TrngleEdge[KK+1][0])-1][1] = 
                          NodeOffset + DipolWid[IEdge]+3;
 
                    GlobalEdgeEnds[abs(TrngleEdge[KK+1][1])-1][0] = 
                          NodeOffset + DipolWid[IEdge]+2;

                    GlobalEdgeEnds[abs(TrngleEdge[KK+1][1])-1][1] =
                          NodeOffset + DipolWid[IEdge]+3;

                }   /* else */ 
    
            }  /* for(j=0 */

        CubeNumNode   += TotDipolNumNode[IEdge];
        CubeNumEdge   += TotDipolNumEdge[IEdge];
        CubeTrngleNum += TotDipolTrngleNum[IEdge];

    }  /* end of ExtcondNum loop */

    /* those external conductive edges */
    for(ii=TotCubeNumEdge;ii<TotEdgeNum;ii++) 
    {  
        EdgeStat[ii]=-1; 
        MetalEdgeStat[ii]=2;
    }   /* for(ii=TotCubeNumEdge */

}       /* end of DiscretizeExterPECPatch() */



/***************************************************************************
Prototype:    void    FindInterMetalEdges() 
Description:  To detect the inner conductive edges 
Input value:     none 
Return value:     none 
Global value used:  CellDimension, Xintcond, Yintcond, Zintcond, NodeCord
Global value modified:     none 
Subroutines called:   none
****************************************************************************/
void FindInterMetalEdges()
{

    int Count, ii;

    for(Count=0;Count<IntcondNum;Count++)
        for(ii=0; ii<IntEsrcNum; ii++)
        {
            int i,j;
            double MinPoint[3], MaxPoint[3];
     
            MinPoint[0]=CellDimension*Xintcond[0][ii]-1e-6;
            MinPoint[1]=CellDimension*Yintcond[0][ii]-1e-6;
            MinPoint[2]=CellDimension*Zintcond[0][ii]-1e-6;
            MaxPoint[0]=CellDimension*Xintcond[1][ii]+1e-6;
            MaxPoint[1]=CellDimension*Yintcond[1][ii]+1e-6;
            MaxPoint[2]=CellDimension*Zintcond[1][ii]+1e-6;
       

            for(i=0; i<TotEdgeNum; i++)
            {
                double start[3], end[3];

                /* detect whether it is within the internal metal area */
                int flag=1;   
       
                /* get coordinates of start and end nodes for one edge 
                   then test whether it is within the metal boundary   */

                for(j=0; j<3; j++) 
                {
                    start[j] = NodeCord[GlobalEdgeEnds[i][0]-1][j];
                    end[j] = NodeCord[GlobalEdgeEnds[i][1]-1][j];
            
            
                    /* not within the boundary, set the flag to 0 */
                    if( (start[j]< MinPoint[j]) || (start[j]>MaxPoint[j] ) ||
                        (end[j]  < MinPoint[j]) || (end[j]  >MaxPoint[j] )  )  
                        flag=0;
                }
 
                /* note: if the edge is within the metal boundary but on the  
                   boundary. do nothing */

                /* the edge is an inner edge and within the metal area, it will 
                   be discarded later*/ 
                if( flag && EdgeStat[i]==1 )  EdgeStat[i]=0;               	    
            } /*for(i=0 */
     
        }   /*  for(ii=0 */
 
}    /* end of FindInterMetalEdges() */




/***************************************************************************
Prototype:    void    DetectISourceEdges() 
Description:  To detect edges that coincides with the current source 
Input value:     none 
Return value:     none 
Global value used:  CellDimension, Xintcond, Yintcond, Zintcond, NodeCord,
    IntEsrcAxis, ISourceEdgeStat, ISourceEdgeMag, 
    TotISourceEdgeNum, IntEsrcMag, 
Global value modified:     none 
Subroutines called:   none
***************************************************************************/
void DetectISourceEdges()
{
    int ii;

    for(ii=0; ii<IntEsrcNum; ii++)
    {
        int i,j;
        double MinPoint[3], MaxPoint[3];
     
        MinPoint[0]=CellDimension*XintEsrc[0][ii]-1e-6;
        MinPoint[1]=CellDimension*YintEsrc[0][ii]-1e-6;
        MinPoint[2]=CellDimension*ZintEsrc[0][ii]-1e-6;
        MaxPoint[0]=CellDimension*XintEsrc[1][ii]+1e-6;
        MaxPoint[1]=CellDimension*YintEsrc[1][ii]+1e-6;
        MaxPoint[2]=CellDimension*ZintEsrc[1][ii]+1e-6;
       

        for(i=0; i<TotEdgeNum; i++)
        {
            double start[3], end[3];
            int flag=1;  /* detect whether it is within the "einter" boundary */
         
            /* get coordinates of start and end nodes for one edge 
               then test whether it is in the Eint boundary */
            for(j=0; j<3; j++) 
            {
                start[j] = NodeCord[GlobalEdgeEnds[i][0]-1][j];
                end[j] = NodeCord[GlobalEdgeEnds[i][1]-1][j];
            
            
                /* not within the boundary, set the flag to false*/
                if((start[j]< MinPoint[j]) || (start[j]>MaxPoint[j] ) ||
                   (end[j]  < MinPoint[j]) || (end[j]  >MaxPoint[j] )  ) flag=0;
            }
         
            /* this edge is not within the source area, skip it */
            if( flag!=1 ) continue;
             
            /* this edge is within the source region,
               thus execute the following codes */
            	    
	    if( IntEsrcAxis[ii]=='x' )
            {
	        /* the edge should be parallel to X axis */
	        if( fabs((double)start[1]-end[1])<1e-6 && 
                    fabs((double)start[2]-end[2])<1e-6 )
                {
                    /* this edge is forcd */
	            ForcdStat[i]=1;  

                    /* store the edge number */
   	            ISourceEdgeStat[TotISourceEdgeNum]=i+1;  
	           
	            /* store the forced value */
	            ISourceEdgeMag[TotISourceEdgeNum]
                         =IntEsrceMag[ii];
	           
	            TotISourceEdgeNum++;  /* increase the index */
	         }    /* if (fabs((double)) */ 
	    }   /* if( 'x' */

	    else if(IntEsrcAxis[ii]=='y')
            {
	        /* the edge should be parallel to Y axis */
	        if( fabs((double)start[0]-end[0])<1e-6 &&
                    fabs((double)start[2]-end[2])<1e-6 )
                {
                    /* this edge is forcd */
	            ForcdStat[i]=1;   

                    /* store the edge number */
	            ISourceEdgeStat[TotISourceEdgeNum]=i+1;	
                 
	            /* store the forced value */
	            ISourceEdgeMag[TotISourceEdgeNum]=
                             IntEsrceMag[ii];
	                 
	            TotISourceEdgeNum++;    /* increase the index */
	        }
	    }

	    else 
            {
	        /* the edge should be parallel to Z axis */
	        if( fabs((double)start[0]-end[0])<1e-6 && 
                    fabs((double)start[1]-end[1])<1e-6)
                {
                    /* this edge is forcd */
	            ForcdStat[i]=1;   	

                    /* store the edge number */
                    ISourceEdgeStat[TotISourceEdgeNum]=i+1;	 
               
	            /* store the forced value */
	            ISourceEdgeMag[TotISourceEdgeNum]=IntEsrceMag[ii];
	            TotISourceEdgeNum++;    /* increase the index */
	        }  /* if( fabs ) */
	    }  /* else */
	}      /* for(i=0)   */
    }          /* for(ii=0)  */


    if (TotISourceEdgeNum==0)
    {
        fprintf(stderr, "Can't find isource edges\n");
        exit(1);
    }
}              /* end of DetectISourceEdges() */
 


/***************************************************************************
Prototype:    void  DetectInnerCond() 
Description:  To detect whether a conductor, or part of the conductor is 
    inside the FEM volume
Input value:     
    Cond pp --- a conductor  
Return value:     none 
Global value used:  XBoundDim, YBoundDim, XBoundDim, Xintcond, Yintcond,
    Zintcond, DelXintcond, DelYintcond, DelZcond, Xextcond, Yextcond,
    Zextcond, IntcondNum, ExtcondNum, 
Global value modified:     none 
Subroutines called:   Intersection()
****************************************************************************/
void DetectInnerCond( Cond pp)
{
    int x1,x2,y1,y2,z1,z2;
    double DelX, DelY, DelZ;
    int i, j;
    RectAngle R1,R2,q,p[5];
 
    x1=pp.x1; x2=pp.x2; 
    y1=pp.y1; y2=pp.y2; 
    z1=pp.z1; z2=pp.z2;
    DelX=pp.DelX;
    DelY=pp.DelY;
    DelZ=pp.DelZ;
           
    /* conductor is parallel to Y-Z axis */
    if ((x1==x2) && ((x1>XBoundDim[0]) || (x1<XBoundDim[1])))
    {
        R1.start.x=YBoundDim[0];  R1.start.y=ZBoundDim[0];
        R1.end.x=YBoundDim[1];    R1.end.y=ZBoundDim[1];
        R2.start.x=y1;            R2.start.y=z1;
        R2.end.x=y2;              R2.end.y=z2;
               
               
        /* find an intersection */                  
        if (Intersection(R1,R2,&q,&i,p))
        {
            /* find a conductative boundary */
            Xintcond[0][IntcondNum]=x1;
            Xintcond[1][IntcondNum]=x1;
            Yintcond[0][IntcondNum]=q.start.x;
            Yintcond[1][IntcondNum]=q.end.x;
            Zintcond[0][IntcondNum]=q.start.y;
            Zintcond[1][IntcondNum]=q.end.y;
            DelXintcond[IntcondNum]= ( int ) DelX;
            DelYintcond[IntcondNum]= ( int ) DelY;
            DelZintcond[IntcondNum]= ( int ) DelZ;
            IntcondNum++;
        }  /* end of if */
               
         /*  sub-areas should be treated as external conductor */
        for(j=0; j<i; j++) 
        {
            Xextcond[0][ExtcondNum]= x1; 
            Xextcond[1][ExtcondNum]= x1;
            Yextcond[0][ExtcondNum]= p[j].start.x; 
            Yextcond[1][ExtcondNum]= p[j].end.x;
            Zextcond[0][ExtcondNum]= p[j].start.y ; 
            Zextcond[1][ExtcondNum]= p[j].end.y;
            DelXextcond[ExtcondNum]= ( int ) DelX;
            DelYextcond[ExtcondNum]= ( int ) DelY;
            DelZextcond[ExtcondNum]= ( int ) DelZ;
            ExtcondNum++;
        } /* end of for */
    }  /* end if */
                  
                  
    /* conductor is parallel to X-Z axis */
    if ((y1==y2)&&((y1==XBoundDim[0])||(y1==XBoundDim[1])))
    {
        R1.start.x=XBoundDim[0]; R1.start.y=ZBoundDim[0];
        R1.end.x=XBoundDim[1];   R1.end.y=ZBoundDim[1];
        R2.start.x=x1;           R2.start.y=z1;
        R2.end.x=x2;             R2.end.y=z2;
               
        /* find an intersection */
        if (Intersection(R1,R2,&q,&i,p))
        {
            /* find a conductative boundary */
            Xintcond[0][IntcondNum]=q.start.x;
            Xintcond[1][IntcondNum]=q.end.x;
            Yintcond[0][IntcondNum]=y1;
            Yintcond[1][IntcondNum]=y1;
            Zintcond[0][IntcondNum]=q.start.y;
            Zintcond[1][IntcondNum]=q.end.y;
            DelXintcond[IntcondNum]= ( int ) DelX;
            DelYintcond[IntcondNum]= ( int ) DelY;
            DelZintcond[IntcondNum]= ( int ) DelZ;
            IntcondNum++;
        }  /* end of if */
               
        /* sub-areas should be treated as external conductors */
        for(j=0; j<i; j++) 
        {
            Xextcond[0][ExtcondNum]= p[j].start.x; 
            Xextcond[1][ExtcondNum]= p[j].end.x;
            Yextcond[0][ExtcondNum]= y1 ; 
            Yextcond[1][ExtcondNum]= y1;
            Zextcond[0][ExtcondNum]= p[j].start.y ; 
            Zextcond[1][ExtcondNum]= p[j].end.y;
            DelXextcond[ExtcondNum]= ( int ) DelX;
            DelYextcond[ExtcondNum]= ( int ) DelY;
            DelZextcond[ExtcondNum]= ( int ) DelZ;
            ExtcondNum++;
        } /* end of for */
    }  /* end if */
                  
                  
    /* conductor is parallel to X-Y axis */
    if ((z1==z2)&&((z1==ZBoundDim[0])||(z1==ZBoundDim[1])))
    {
        R1.start.x=XBoundDim[0]; R1.start.y=YBoundDim[0];
        R1.end.x=XBoundDim[1];   R1.end.y=YBoundDim[1];
        R2.start.x=x1;           R2.start.y=y1;
        R2.end.x=x2;             R2.end.y=y2;
               
        if (Intersection(R1,R2,&q,&i,p))
        {
            /* find a conductative boundary */
            Xintcond[0][IntcondNum]=q.start.x;
            Xintcond[1][IntcondNum]=q.end.x;
            Yintcond[0][IntcondNum]=q.start.y;
            Yintcond[1][IntcondNum]=q.end.y;
            Zintcond[0][IntcondNum]=z1;
            Zintcond[1][IntcondNum]=z1;
            DelXintcond[IntcondNum]= ( int ) DelX;
            DelYintcond[IntcondNum]= ( int ) DelY;
            DelZintcond[IntcondNum]= ( int ) DelZ;
            IntcondNum++;
        }  /* end of if */
               
               
        for(j=0; j<i; j++) 
        {
            Xextcond[0][ExtcondNum]= p[j].start.x; 
            Xextcond[1][ExtcondNum]= p[j].end.x;               
            Yextcond[0][ExtcondNum]= p[j].start.y ; 
            Yextcond[1][ExtcondNum]= p[j].end.y;
            Zextcond[0][ExtcondNum]= z1 ; 
            Zextcond[1][ExtcondNum]= z1;
            DelXextcond[ExtcondNum]= ( int ) DelX;
            DelYextcond[ExtcondNum]= ( int ) DelY;
            DelZextcond[ExtcondNum]= ( int ) DelZ;
            ExtcondNum++;
        } /* end of for */
    }  /* end if */
               
}


/***************************************************************************
Prototype:    int  DetectCondBound() 
Description: To detect whether a conductor or part of a conductor is 
    located on the FEM surface. 
Input value:    
    Cond pp --- a conductor.
Return value:    If ture, return 1, otherwise return 0  none 
Global value used:  XBoundDim, YBoundDim, ZBoundDim, Xintcond, Yintcond,
    Zintcond, DelXintcond, DelYintcond, DelZintcond, Xextcond, Yextcond,
    Zextcond, IntcondNum, ExtcondNum, Xbndcond, Ybndcond, Zbndcond,
    DelXbndcond, DelYbndcond, DelZbndcond, 
Global value modified:     none 
Subroutines called:   Inetersection()
***************************************************************************/
int DetectCondBound( Cond pp)
{

    int x1,x2,y1,y2,z1,z2;
    double DelX, DelY, DelZ;
    int i, j,flag;
    RectAngle R1,R2,q,p[5];
 
    x1=pp.x1;   x2=pp.x2; 
    y1=pp.y1; y2=pp.y2; 
    z1=pp.z1; z2=pp.z2;
    DelX=pp.DelX;
    DelY=pp.DelY;
    DelZ=pp.DelZ;

  
    flag=0;

           
    /* conductor is parallel to Y-Z axis */
    if ((x1==x2) && ((x1==XBoundDim[0]) || (x1==XBoundDim[1])))
    {
        R1.start.x=YBoundDim[0];  
        R1.start.y=ZBoundDim[0];
        R1.end.x=YBoundDim[1];    
        R1.end.y=ZBoundDim[1];
        R2.start.x=y1;            
        R2.start.y=z1;
        R2.end.x=y2;              
        R2.end.y=z2;
               
               
        /* find an intersection */
        if (Intersection(R1,R2,&q,&i,p))
        {
            /* find a conductative boundary */
                   
            flag=1;
            Xbndcond[0][BndcondNum]=x1;
            Xbndcond[1][BndcondNum]=x1;
            Ybndcond[0][BndcondNum]=q.start.x;
            Ybndcond[1][BndcondNum]=q.end.x;
            Zbndcond[0][BndcondNum]=q.start.y;
            Zbndcond[1][BndcondNum]=q.end.y;
            DelXbndcond[BndcondNum]= ( int ) DelX;
            DelYbndcond[BndcondNum]= ( int ) DelY;
            DelZbndcond[BndcondNum]= ( int ) DelZ;
            BndcondNum++;
        }  /* end of if */
               
        /*  sub-areas should be treated as external conductor */
        for(j=0; j<i; j++) 
        {
            Xextcond[0][ExtcondNum]= x1; 
            Xextcond[1][ExtcondNum]= x1;
            Yextcond[0][ExtcondNum]= p[j].start.x; 
            Yextcond[1][ExtcondNum]= p[j].end.x;
            Zextcond[0][ExtcondNum]= p[j].start.y ; 
            Zextcond[1][ExtcondNum]= p[j].end.y;
            DelXextcond[ExtcondNum]= ( int ) DelX;
            DelYextcond[ExtcondNum]= ( int ) DelY;
            DelZextcond[ExtcondNum]= ( int ) DelZ;
            ExtcondNum++;
        } /* end of for */
    }  /* end if */
                  

    /* conductor is parallel to X-Z axis */
    if ((y1==y2)&&((y1==YBoundDim[0])||(y1==YBoundDim[1])))
    {
        R1.start.x=XBoundDim[0]; 
        R1.start.y=ZBoundDim[0];
        R1.end.x=XBoundDim[1];
        R1.end.y=ZBoundDim[1];
        R2.start.x=x1;          
        R2.start.y=z1;
        R2.end.x=x2;             
        R2.end.y=z2;
               
        /* find an intersection */
        if (Intersection(R1,R2,&q,&i,p))
        {
               
            /* find a conductative boundary */
            flag=1;
            Xbndcond[0][BndcondNum]=q.start.x;
            Xbndcond[1][BndcondNum]=q.end.x;
            Ybndcond[0][BndcondNum]=y1;
            Ybndcond[1][BndcondNum]=y1;
            Zbndcond[0][BndcondNum]=q.start.y;
            Zbndcond[1][BndcondNum]=q.end.y;
            DelXbndcond[BndcondNum]= ( int ) DelX;
            DelYbndcond[BndcondNum]= ( int ) DelY;
            DelZbndcond[BndcondNum]= ( int ) DelZ;
            BndcondNum++;
        }  /* end of if */
               
        /* sub-areas should be treated as external conductors */
        for(j=0; j<i; j++) 
        {
            Xextcond[0][ExtcondNum]= p[j].start.x; 
            Xextcond[1][ExtcondNum]= p[j].end.x;
            Yextcond[0][ExtcondNum]= y1 ; 
            Yextcond[1][ExtcondNum]= y1;
            Zextcond[0][ExtcondNum]= p[j].start.y ; 
            Zextcond[1][ExtcondNum]= p[j].end.y;
            DelXextcond[ExtcondNum]= ( int ) DelX;
            DelYextcond[ExtcondNum]= ( int ) DelY;
            DelZextcond[ExtcondNum]= ( int ) DelZ;
            ExtcondNum++;
        } /* end of for */
    }  /* end if */
                  
                  
    /* conductor is parallel to X-Y axis */
    if ((z1==z2)&&((z1==ZBoundDim[0])||(z1==ZBoundDim[1])))
    {
        R1.start.x=XBoundDim[0]; R1.start.y=YBoundDim[0];
        R1.end.x=XBoundDim[1];   R1.end.y=YBoundDim[1];
        R2.start.x=x1;           R2.start.y=y1;
        R2.end.x=x2;             R2.end.y=y2;
               
               
        if (Intersection(R1,R2,&q,&i,p))
        {
            /* find a conductative boundary */
            flag=1;
            Xbndcond[0][BndcondNum]=q.start.x;
            Xbndcond[1][BndcondNum]=q.end.x;
            Ybndcond[0][BndcondNum]=q.start.y;
            Ybndcond[1][BndcondNum]=q.end.y;
            Zbndcond[0][BndcondNum]=z1;
            Zbndcond[1][BndcondNum]=z1;
            DelXbndcond[BndcondNum]= ( int ) DelX;
            DelYbndcond[BndcondNum]= ( int ) DelY;
            DelZbndcond[BndcondNum]= ( int ) DelZ;
            BndcondNum++;
        }  /* end of if */
               
        for(j=0; j<i; j++) 
        {
            Xextcond[0][ExtcondNum]= p[j].start.x; 
            Xextcond[1][ExtcondNum]= p[j].end.x;
            Yextcond[0][ExtcondNum]= p[j].start.y ; 
            Yextcond[1][ExtcondNum]= p[j].end.y;
            Zextcond[0][ExtcondNum]= z1 ; 
            Zextcond[1][ExtcondNum]= z1;
            DelXextcond[ExtcondNum]= ( int ) DelX;
            DelYextcond[ExtcondNum]= ( int ) DelY;
            DelZextcond[ExtcondNum]= ( int ) DelZ;
            ExtcondNum++;
        } /* end of for */
    }  /* end if */
           
           
    return flag;
               
}  /* end of DetectCondBound()  */




/*****************  END OF CODE  ***********************/
